/* Matthew Phillips mephilli 20066500 CS488 A5
 *
 * Copyright (c) 2005 Matthew Phillips
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/** @file battleship_logic.c 
 *
 * Battleship logic.
 * 
 * These files are responsible for the majority of the game logic. (Some game
 * logic is handled by battleship_menu.c.)
 */
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>

#include "battleship_logic.h"
#include "battleship_view.h"
#include "battleship_model.h"
#include "battleship_menu.h"
#include "battleship_rocket.h"
#include "battleship_sound.h"
#include "glfont.h"

/** Delay used for animation call back */
#define ANIMATION_SPEED 55
/** Delay used for AI call back */
#define AI_SPEED 100
/** Addition timer delay for demo mode */
#define DEMO_DELAY 20

/** The final size of message text */
#define MAX_MSG_SIZE 5
/** The initial size of message text */
#define MIN_MSG_SIZE 1
/** How long to leave message text persistant */
#define MAX_MSG_DELAY 35
/** The starting delay value for hit/miss messages*/
#define HIT_MSG_DELAY (MAX_MSG_DELAY - 6)
/** The starting delay value for sunk messages */
#define SUNK_MSG_DELAY (MAX_MSG_DELAY - 10)
/** The starting delay value for game over message */
#define WIN_MSG_DELAY 0

/** Quick definition of next player */
#define next_player 1 - curr_player

/** Keeps track of which mouse buttons are pressed */
static unsigned int mouse_button;
/** Keeps track of key modifiers used with the mouse */
static unsigned int mouse_mod;
/** Keeps track of last mouse location */
static int mouse_x;
/** Keeps track of last mouse location */
static int mouse_y;

/** Attempts to place a ship. 
 *  
 *  The value of curr_player is used to index into the player array. And the
 *  given value is used to index into that players ship array. The function
 *  then determines if the ship can be placed at its current location. If
 *  it can it places it and sets its state to placed.
 *
 *  @param num The ship number to try and place.
 *  @return Returns true if the ship is placed succesfully
 */ 
static bool try_place_ship(unsigned char num);

/** Fires at the currently selected location.
 *
 * 	The curr_player variable is used to index the player array. The value
 * 	of sel_x and sel_y for that player are used to determine here to fire.
 * 	This function determines if there is a hit or a misses and updates
 * 	all the required variables.
 *
 * 	@return The function returns true if the game is over. Otherwise false.
 */
static bool fire();

/** Moves the current players selection intelligantly.
 *
 *  The players cursor is moved by xm, ym. If the player
 *  is no an ai then blocks which have pegs are skipped.
 *
 *  @param xm How much to move left/right (positive is right)
 *  @param ym How much to move up/down (positive is down)
 */
static void fix_select(char xm, char ym);

/** Moves the game to the next state */
static void next_state();

/** Switchs to the given player
 *
 *  Changes the curr_player variable to pnum.
 *  It also initiates the graphical borad rotation so that
 *  the new player will be able to see their side of the board.
 * 
 *  @param pnum The player to swtich to.
 */
static void switch_player(unsigned char pnum);

/** Automatically places the current players ships
 *
 *  Called when a players auto_place variable is set to true. The
 *  currenty players ships are placed randomly on the board.
 */
static void autoplace_ships();

/** Used to keep track of information about what ships an AI has hi. 
 *
 *  When an AI hits a ship it tries to sink it by moveing one space at
 *  a time in a direction. It continues to move in that direction as
 *  long as it hits things. When it stops hitting things, and it has not
 *  sunk the ship it returns to the origin and tries a new direction.
 */
struct ai_hit
{
	/** x origin of the hit */
	unsigned char x; 
	/** y oringin of the hit */
	unsigned char y;
	/** Which directions have been tried.
	 *  When a direction has been tired it is set to true. The
	 *  indexes corrisond to up, right, down, left.
	 */
	bool dir[4];
	/** The current direciton that is being attempted */
	unsigned char curr_dir;
};

/** Used to keep track of AI movements.
 *  @see struct ai_hit
 */
struct ai_move
{
	/** x location the AI will fire at */
	unsigned char x;
	/** y location the AI will fire at */	
	unsigned char y;
	/** True if the AI has decided where to fire */
	bool chosen;
	/** Used for hard AI. Eiterh 0 or 1. Determines where the start
	 *  of the shooting grid is
	 */
	unsigned char mod;
	/** Used for the hard AI. Between 0 and 4. 
	 *  @see ai_pick_move()
	 */
	unsigned char state;
	/** For keeping track of which ships have been hit */
	struct ai_hit hit[NUM_SHIPS];
	/** How many ships have been hit (and not sunk) */
	unsigned char num_hits;
};

/** One each to keep track of AI movements */
static struct ai_move ai_move[2];

/** Causes the AI to play. 
 *
 *  This functions is like the animation functions. It is called repeatedly
 *  by a timer. It first decided where the AI will play next. (if the AI
 *  hasn't already decided. This happends when a hit has occured). Then
 *  it moves square by square (one call back for each square) to the location.
 *  Then it fires. 
 *
 *  If a key is pressed during the movement, the animation is cancled and
 *  the AI jumps right to the fireing locaiton and fires.
 */
static bool ai_play();

/** Picks a location for the AI to fire.
 *  See code for details.
 */
static void ai_pick_move();

/** Performes AI calcuations for when a ship is hit.
 *  See code for details.
 */
static void ai_follow_hit();

/** Descirbes current status of borad movement. 
 *
 *  The board is rotated and moved during game play. Sometimes
 *  in multiple way at the same time. This struct is used to
 *  keep track of everything so that all motion can be handled
 *  with a single animation callback/timer.
 */
struct board_motion
{
	/** Rate of change of rotation for a given axis */
	GLfloat rrate[3];
	/** Maximum rotation for given axis */
	GLfloat rend[3];
	/** Rate of change of translation for a given axis */
	GLfloat trate[3];
	/** Maximum translation for given axis */
	GLfloat tend[3];
};

/** Descirbes current status of borad movement. */
static struct board_motion board_motion;

/** Initiates an animated board movement.
 * 
 *	abs(currently location - to) should be divisable by rate.
 * 
 *  @param axis The axis to move on
 *  @param to   Where on the axis to move to
 *  @param rate How much to move to move by each timer.
 */
static void board_move(size_t axis, GLfloat to, GLfloat rate);

/** Initiates an animated board rotation.
 * 
 *	abs(currently rotation - to) should be divisable by rate.
 * 
 *  @param axis The axis to rotate on
 *  @param to   What angle to rotate to
 *  @param rate How much to rotate by each timer.
 */
static void board_rotate(size_t axis, GLfloat to, GLfloat rate);

/** Sets up the timer for a board movement animation */
static void start_transition();
/** Sets up the timer for a AI animation */
static void start_ai();
/** Sets up the timer for a fireing animation */
static void start_fireing();
/** Sets up the timer for a message animation */
static void start_message();

/** This variable is set to true to tell the current animation to finsih. */
static bool cancel_animation;

/** Called by a timer for board animation */
static bool animate_transition();
/** Called by a timer for firesing animation */
static bool animate_fireing();
/** Called by a timer for message animation */
static bool animate_message();

/** Store that the game is over */
static bool game_over;

void
battleship_logic_init()
{
	static bool first_time = true;
	battleship_model_init();
	battleship_view_init();
	battleship_menu_init();
	if (first_time)
	{
		battleship_sound_init();
		first_time = false;
	}
	mouse_button = 0;
	
	size_t i;
	for (i = 0; i < 3; i++)
	{
		board_motion.rrate[i] = 0;
		board_motion.rend[i] = 0;
		board_motion.trate[i] = 0;
		board_motion.tend[i] = 0;
	}

	cancel_animation = false;
	for (i = 0; i < 2; i++)
	{
		ai_move[i].chosen = false;
		ai_move[i].mod = (rand()/(RAND_MAX+1.0) > 0.5);
		ai_move[i].state = 0;
		ai_move[i].num_hits = 0;
	}

	game_over = false;
	game_state = GAME_INIT;
	if (demo_mode)
		battleship_logic_demo();
	else		
		battleship_menu_load(&main_menu);
}

void
battleship_logic_new()
{
	next_state();	
}

void 
battleship_logic_onep()
{
	player[0].ai = AI_HUMAN;
	player[0].auto_place = false;
	player[0].fog = false;
	player[1].ai = AI_HARD;
	player[1].auto_place = true;
	player[1].fog = true;

	battleship_logic_new();
}

void 
battleship_logic_twop()
{
	player[0].ai = AI_HUMAN;
	player[0].auto_place = false;
	player[0].fog = true;
	player[1].ai = AI_HUMAN;
	player[1].auto_place = false;
	player[1].fog = true;

	battleship_logic_new();
}

void
battleship_logic_demo()
{
	demo_mode = true;
	
	player[0].ai = ((double) 3.0*rand()/(RAND_MAX+1.0)) + 1;
	player[0].auto_place = true;
	player[0].fog = false;
	player[1].ai = ((double) 3.0*rand()/(RAND_MAX+1.0)) + 1;
	player[1].auto_place = true;
	player[1].fog = false;
	
	battleship_logic_new();
}

void
battleship_logic_restart()
{
	battleship_logic_stop_timers();
	battleship_logic_init();
	battleship_sound_stop();
	battleship_view_refresh();
}

void 
battleship_logic_mouse_move(int x, int y)
{
	int w, h;
	battleship_view_getsize(&w, &h);
	float diffx = (x - mouse_x) / (float) w;
	float diffy = (y - mouse_y) / (float) h;
	
	bool needRefresh = false;
	if (mouse_mod & BATTLESHIP_MOD_SHIFT)
	{
		float tdiffx = diffx * 15.0;
		float tdiffy = diffy * 15.0;
		if (mouse_button & BATTLESHIP_MOUSE_LEFT)
		{
			battleship_view_set_translate(0, tdiffx, 'U');
		}
		if (mouse_button & BATTLESHIP_MOUSE_MIDDLE)
		{
			battleship_view_set_translate(1, -tdiffy, 'U');
		}
		if (mouse_button & BATTLESHIP_MOUSE_RIGHT)
		{
			battleship_view_set_translate(2, tdiffy, 'U');
		}
		needRefresh = true;
	}
	if (mouse_mod & BATTLESHIP_MOD_CRTL)
	{
		float rdiffx = diffx * 360;
		float rdiffy = diffy * 360;
		if (mouse_button & BATTLESHIP_MOUSE_LEFT)
		{
			battleship_view_set_rotate(1, rdiffx, 'U');
		}
		if (mouse_button & BATTLESHIP_MOUSE_MIDDLE)
		{
			battleship_view_set_rotate(0, rdiffy, 'U');
		}
		if (mouse_button & BATTLESHIP_MOUSE_RIGHT)
		{
			battleship_view_set_rotate(2, -rdiffx, 'U');
		}
		needRefresh = true;
	}
	
	mouse_x = x;
	mouse_y = y;

	if (needRefresh)
		battleship_view_refresh();
}

void 
battleship_logic_mouse_up(unsigned int button, unsigned int mod, int x, int y)
{
	//NOTE: for the time being don't care about these
	(void) x;
	(void) y;

	mouse_button ^= (1 << (button - 1));
	mouse_mod = mod;
}

void 
battleship_logic_mouse_down(unsigned int button, unsigned int mod, int x, int y)
{
	mouse_x = x; 
	mouse_y = y;
	mouse_button |= (1 << (button - 1));
	mouse_mod = mod;
}

void
battleship_logic_keypress(unsigned char c, unsigned int mod)
{
	bool needRefresh = true;

	if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_QUIT])
		battleship_logic_quit();

	if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_RESET])
	{
		size_t i = 0;
		for (i = 0; i < 3; i++)
		{
			battleship_view_set_translate(i, 0.0, 'u');
			battleship_view_set_rotate(i, 0.0, 'u');
		}
		battleship_view_refresh();
	}
	
	if (curr_menu)
	{
		if (name_selector.enabled)
		{
			if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_UP])
				name_selector.y = (GLFONT_TEST_HEIGHT + name_selector.y - 1) % GLFONT_TEST_HEIGHT;
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_DOWN])
				name_selector.y = (name_selector.y + 1) % GLFONT_TEST_HEIGHT;
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_LEFT])
				name_selector.x = (GLFONT_TEST_WIDTH + name_selector.x - 1) % GLFONT_TEST_WIDTH;
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_RIGHT])
				name_selector.x = (name_selector.x + 1) % GLFONT_TEST_WIDTH;
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_ENTER])
			{
				size_t len = strlen(name_selector.name);
				if (len < MAX_NAME_LEN)
				{
					name_selector.name[len] = glfont_test_char(name_selector.x,
															  name_selector.y);
					name_selector.name[len + 1] = '\0';
				}
				else
					needRefresh = false;
			}
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_ESC])
			{
				name_selector.enabled = false;
			}
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_BACKSPACE])
			{
				size_t len = strlen(name_selector.name);
				if (len > 0)
					name_selector.name[len - 1] = '\0';
				else
					needRefresh = false;
			}
			else
			{
				size_t len = strlen(name_selector.name);
				int x = 0; int y = 0;
				needRefresh = glfont_is_test_char(c,&x,&y)&&(len<MAX_NAME_LEN);
				if (needRefresh)
				{
					name_selector.name[len] = c; 
					name_selector.name[len + 1] = '\0';
					name_selector.x = x;
					name_selector.y = y;
				}
			}
		}
		else
		{
			if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_UP])
				battleship_menu_up();
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_DOWN])
				battleship_menu_down();
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_LEFT])
				battleship_menu_left();
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_RIGHT])
				battleship_menu_right();
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_ENTER])
				battleship_menu_toggle();
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_ESC])
			{
				if (game_state != GAME_INIT || curr_menu->parent)
				{
					battleship_menu_cancel();
					if (!curr_menu && game_state == GAME_PLAYING)
					{
						//if a player was previously an ai player need
						//to clear what it knows
						size_t i;
						for (i = 0; i < 2; i++)
						{
							if (!player[i].ai)
							{
								ai_move[i].chosen = false;
								ai_move[i].num_hits = 0;
							}
						}
					
						//restart ai player
						if (player[curr_player].ai)
							start_ai();
					}
				}
				else
					needRefresh = false;
			}
			else 
				needRefresh = false;
			
		}	
		
		if (needRefresh)
			battleship_view_refresh();

		return;
	}
	
	if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_ESC])
	{
		//stop demo mode on ecp
		if (demo_mode)
		{
			demo_mode = false;
			battleship_logic_restart();
			return;
		}
		//options menu
		else if (!curr_menu)
		{
			battleship_menu_load(&options_menu);
			battleship_view_refresh();
			return;
		}
	}

	switch (game_state)
	{
		case GAME_PLACE_SHIPS:
			//you aren't allowed to declair variables as the first thing
			//in a switch. One day maybe someone will tell me why
			(void) game_state;
			
			//find current ship
			struct battleship_ship *curr_ship = NULL;
			unsigned char curr_ship_t = 0;
			size_t i;
			for (i = 0; i < NUM_SHIPS; i++)
			{
				if (player[curr_player].ship[i].state == SHIP_PLACING)
				{
					curr_ship = &player[curr_player].ship[i];
					curr_ship_t = i;
					break;
				}
			}
			if (!curr_ship) return;
			unsigned char hl = (curr_ship->down) ? 1 : curr_ship->length;
			unsigned char vl = (!curr_ship->down) ? 1 : curr_ship->length;
			
			if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_LEFT])
			{
				if (curr_ship->x > 0)
					--(curr_ship->x);
			}
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_RIGHT])
			{
				if (curr_ship->x + hl < GRID_DIM)
					++(curr_ship->x);
			}
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_UP])
			{
				if (curr_ship->y > 0)
					--(curr_ship->y);
			}
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_DOWN])
			{
				if (curr_ship->y + vl < GRID_DIM)
					++(curr_ship->y);
			}
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_ENTER])
			{
				if (try_place_ship(curr_ship_t))
				{
					//If the ship is placed and it is the last one move on
					if (curr_ship_t == NUM_SHIPS - 1)
						next_state();
					else
						player[curr_player].ship[curr_ship_t + 1].state = SHIP_PLACING;
				}
			}
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_BACKSPACE])
			{
				if (curr_ship_t != 0)
				{
					//Stop placing current ship
					curr_ship->state = SHIP_NOT_PLACED;
					//Switch to last ship
					curr_ship = &player[curr_player].ship[curr_ship_t - 1];
					hl = (curr_ship->down) ? 1 : curr_ship->length;
					vl = (!curr_ship->down) ? 1 : curr_ship->length;
					
					//Unplace it
					size_t j;
					for (i = 0; i < hl; i++)
					{
						for (j = 0; j < vl; j++)
						{
							player[curr_player].ship_at
								[curr_ship->x + i][curr_ship->y +j] = NO_SHIP; 
						}
					}

					//Set it to placing
					curr_ship->state = SHIP_PLACING;
				}
			}
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_ROTATE])
			{
				curr_ship->down = !curr_ship->down;
			}
			else
			{
				needRefresh = false;
			}
		break;
		case GAME_TRANSITION:
		case GAME_FIREING:
		case GAME_MESSAGE:
			cancel_animation = true;
		break;
		case GAME_AI_PLAYING:
			if (mod != BATTLESHIP_MOD_AI) 
			{
				cancel_animation = true;;
				return;
			}
		//note lack of break	
		case GAME_PLAYING:
			if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_LEFT])
			{
				fix_select(-1, 0);
			}
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_RIGHT])
			{
				fix_select(1, 0);
			}
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_UP])
			{
				fix_select(0, -1);
			}
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_DOWN])
			{
				fix_select(0, 1);
			}
			else if (c == BATTLESHIP_KEYS[BATTLESHIP_KEY_ENTER])
			{
				if (player[next_player].grid
						[player[curr_player].sel_x][player[curr_player].sel_y] 
						== GRID_EMPTY)
				{
					next_state();
					needRefresh = false;
				}
			}
			else
			{
				needRefresh = false;
			}
		break;
		case GAME_OVER:
			next_state();
		break;
		default:
			needRefresh = false;
		break;
	}


	if (needRefresh)
		battleship_view_refresh();
}

static void
next_state()
{
	switch (game_state)
	{
		case GAME_INIT: 
			curr_menu = NULL;
			game_state = GAME_PLACE_SHIPS;
			
			//put first ship in placing state
			player[0].ship[0].state = SHIP_PLACING;
			player[1].ship[0].state = SHIP_PLACING;
			
			//do ai placing of ships
			curr_player = 0;
			if (player[0].auto_place) autoplace_ships();
			curr_player = 1;
			if (player[1].auto_place) autoplace_ships();
			curr_player = 0;

			//Don't go to ship placing state if both players are ai's
			if (player[0].auto_place && player[1].auto_place)
			{
				game_state = GAME_PLAYING;
				board_rotate(0, 15.0, 10.0);
				board_move(1, 0.0, 0.5);
				if (player[0].ai)
					start_ai();
			}
			else
			{
				//Move to next player if first players is AI
				if (player[0].auto_place)
					switch_player(next_player);
				board_rotate(0, 45.0, 5.0);
				board_move(1, 4.5, 0.5);
			}
		break;	
		case GAME_PLACE_SHIPS:
			if (curr_player == 0 && !player[1].auto_place)
			{
				switch_player(next_player);
			}
			else
			{
				game_state = GAME_PLAYING;
				board_rotate(0, 15.0, 10.0);
				board_move(1, 0.0, 0.5);
				switch_player(0);
				if (player[0].ai)
					start_ai();
			}
		break;	
		case GAME_AI_PLAYING:
		case GAME_PLAYING:
			if (game_rocket.type != ROCKET_OFF)
				board_rotate(1, 90, 10.0);
			battleship_view_ready_rocket();
			start_fireing();
		break;	
		case GAME_FIREING:
			//fire and dtermine if the game is over
			game_over = fire();
			start_message();
			
			//don't start sound if animation is not enabled, it will just
			//be cancled
			if (do_msg_animation)
			{
				if (game_message.type == GRID_MISS)
					battleship_sound_play(BATTLESHIP_SOUND_MISS);
				else if (game_message.type == GRID_HIT)
				{
					if (game_message.ship)
						battleship_sound_play(BATTLESHIP_SOUND_SUNK);
					else
						battleship_sound_play(BATTLESHIP_SOUND_HIT);
				}
				else if (game_message.type == GRID_EMPTY)
					battleship_sound_play(BATTLESHIP_SOUND_WIN);
			}
		break;
		case GAME_MESSAGE:
			//if the game is over don't continue
			if (game_over)
			{
				//If demo mode is active restart automatically
				if (game_over && demo_mode)
				{
					battleship_logic_restart();
				}
				else
					game_state = GAME_OVER;
			}
			//back to playing
			else
			{
				game_state = GAME_PLAYING;
				if (!player[curr_player].ai)
					fix_select(1, 0);
				else if (player[curr_player].ai != AI_EASY)
					ai_follow_hit();
				switch_player(next_player);
				if (player[curr_player].ai)
					start_ai();
			}
		break;	
		case GAME_OVER:
			battleship_logic_restart();
		break;
		default:
			//DPRINT("Error: Unknown or unexpected state %i\n", game_state);
		break;
	}
}

static bool
try_place_ship(unsigned char num)
{
	struct battleship_ship *curr_ship = &player[curr_player].ship[num];
	unsigned char hl = (curr_ship->down) ? 1 : curr_ship->length;
	unsigned char vl = (!curr_ship->down) ? 1 : curr_ship->length;

	//check for ship
	bool conflict = false;
	size_t i, j;
	for (i = 0; !conflict && i < hl; i++)
	{
		for (j = 0; !conflict && j < vl; j++)
		{
			if (curr_ship->y + j >= GRID_DIM || curr_ship->x + i >= GRID_DIM ||	
				player[curr_player].ship_at
					[curr_ship->x + i][curr_ship->y +j]	!= NO_SHIP 
				)
			{
				conflict = true;
			}
		}
	}

	if (!conflict)
	{
		for (i = 0; i < hl; i++)
		{
			for (j = 0; j < vl; j++)
			{
				player[curr_player].ship_at
					[curr_ship->x + i][curr_ship->y +j] = num; 
			}
		}
		curr_ship->state = SHIP_PLACED;
	}

	return !conflict;
}

static bool
fire()
{
	bool go = false;
	game_message.size = MIN_MSG_SIZE;
	game_message.delay = HIT_MSG_DELAY;
	game_message.ship = NULL;
	
	unsigned char sel_x = player[curr_player].sel_x;
	unsigned char sel_y = player[curr_player].sel_y;
	if (player[next_player].ship_at[sel_x][sel_y] == NO_SHIP)
	{
		game_message.type = GRID_MISS;

		player[next_player].grid[sel_x][sel_y] = GRID_MISS;
		++player[curr_player].num_misses;
	}
	else
	{
		game_message.type = GRID_HIT;

		//add hit
		player[next_player].grid[sel_x][sel_y] = GRID_HIT;
		++player[curr_player].num_hits;
		
		//add hit to ship
		unsigned char ship=player[next_player].ship_at[sel_x][sel_y];
		++player[next_player].ship[ship].num_hits;

		//shink
		if (player[next_player].ship[ship].num_hits == player[next_player].ship[ship].length)
		{
			game_message.ship = (char *) battleship_shiptype_s[ship];
			game_message.delay = SUNK_MSG_DELAY;

			player[next_player].ship[ship].state = SHIP_SUNK;
			++player[next_player].num_sunk;
			if (player[next_player].num_sunk == NUM_SHIPS)
			{
				game_message.delay = WIN_MSG_DELAY;
				game_message.ship = NULL;
				game_message.type = GRID_EMPTY;
				go = true;
			}
		}
	}

	return go;
}

static void
fix_select(char xm, char ym)
{
	player[curr_player].sel_x = 
		(player[curr_player].sel_x + (xm + GRID_DIM)) % GRID_DIM;
	player[curr_player].sel_y = 
		(player[curr_player].sel_y + (ym + GRID_DIM)) % GRID_DIM;
}

static void
switch_player(unsigned char pnum)
{
	if (curr_player != pnum)
	{
		curr_player = pnum;
		board_rotate(1, (curr_player == 0) ? 0 : 180, 10.0);
	}
}


static bool
ai_play()
{
	//Don't play while menu is open
	//Its also possible the user will change from AI to human using options menu
	if (curr_menu || !player[curr_player].ai)
	{
		cancel_animation = false;
		game_state = GAME_PLAYING;
		return false;
	}
	
	struct ai_move *ai = &ai_move[curr_player];
	bool r = true;

	//Find a new place to shoot
	if (!ai->chosen)
		ai_pick_move();

	//move there right away if animation is off
	if (cancel_animation || !do_ai_animation) 
	{
		player[curr_player].sel_x = ai->x;
		player[curr_player].sel_y = ai->y;
	}
	
	//A place has been chosen and the cursor is on it. Fire.
	if (player[curr_player].sel_x == ai->x &&
			 player[curr_player].sel_y == ai->y
			 )
	{
		//this needs to come before hit checking code. because hit checking
		//code may correctly set it back to true
		ai->chosen = false;
		cancel_animation = false;
		battleship_logic_keypress(
				BATTLESHIP_KEYS[BATTLESHIP_KEY_ENTER],BATTLESHIP_MOD_AI);
		r = false;
	}
	//A place has been chosen but the cursor is not on it. So move.
	else
	{
		//move to location
        if (player[curr_player].sel_x < ai->x)
            battleship_logic_keypress(
					BATTLESHIP_KEYS[BATTLESHIP_KEY_RIGHT], BATTLESHIP_MOD_AI);
        else if (player[curr_player].sel_x > ai->x)
            battleship_logic_keypress(
					BATTLESHIP_KEYS[BATTLESHIP_KEY_LEFT], BATTLESHIP_MOD_AI);
        else if (player[curr_player].sel_y < ai->y)
            battleship_logic_keypress(
					BATTLESHIP_KEYS[BATTLESHIP_KEY_DOWN], BATTLESHIP_MOD_AI);
        else if (player[curr_player].sel_y > ai->y)
            battleship_logic_keypress(
					BATTLESHIP_KEYS[BATTLESHIP_KEY_UP], BATTLESHIP_MOD_AI);
	}

	return r;
}

static void
ai_pick_move()
{
	struct ai_move *ai = &ai_move[curr_player];
	ai->x = (unsigned char) ((double) GRID_DIM*rand()/(RAND_MAX+1.0));
	ai->y = (unsigned char) ((double) GRID_DIM*rand()/(RAND_MAX+1.0));
	unsigned char initx = ai->x; 
	unsigned char inity = ai->y;
	while (true)
	{
		if (player[next_player].grid[ai->x][ai->y] == GRID_EMPTY)
		{
			//Only hard AI does useful picking.
			//Technically AI state should never reach 4
			if (player[curr_player].ai != AI_HARD || ai->state == 3) 
				break;

			//the ai has narrowed the board down to open blocks of 2
			bool left = 
			(ai->x == 0) ||
			(player[next_player].grid[ai->x - 1][ai->y] != GRID_EMPTY);
			bool right = 
			(ai->x == GRID_DIM - 1) || 
			(player[next_player].grid[ai->x + 1][ai->y] != GRID_EMPTY);
			bool up = 
			(ai->y == 0) || 
			(player[next_player].grid[ai->x][ai->y - 1] != GRID_EMPTY);
			bool down = 
			(ai->y == GRID_DIM - 1) || 
			(player[next_player].grid[ai->x][ai->y + 1] != GRID_EMPTY);
			
			int open = 4 - (up + down + left + right);
		
			//Ensures 2 space gap between pieces (but don't pick a place
			//that has fewer than 3 open)
			if (ai->state == 0 && ((ai->y + ai->mod) % 3 == ai->x % 3)
					&& open > 2)
				break;	
			//Try all spaces that have 2 open space around them
			else if (ai->state == 1 && open > 1)
				break;
			//Now try all space that have one open space
			else if (ai->state == 2 && open > 0)
				break;
		}

		++(ai->x);
		if (ai->x >= GRID_DIM)
		{
			++(ai->y);
			ai->x = 0;
			if (ai->y >= GRID_DIM)
				ai->y = 0;
		}

		//The hard AI has 3 search states. It determines that it needs
		//to move to the next when there are no moves left for the one its
		//on.
		if (ai->x == initx && ai->y == inity)
		{
			if (ai->state == 4 || player[curr_player].ai != AI_HARD)
			{
				//DPRINT("Error: Board is full\n");
				return;
			}
			++ai->state;
		}
	}
	ai->chosen = true;
}

static void
ai_follow_hit()
{
	struct ai_move *ai = &ai_move[curr_player];

	enum battleship_shiptype ship = player[next_player].ship_at[ai->x][ai->y];
	
	if (ship != NO_SHIP)
	{
		//This loop both checks the hit queue for sunk ships and
		//determines if the ship that has been hit is in the queue
		bool new = true;
		enum battleship_shiptype ship2;
		size_t i = 0;
		for ( ; ; )
		{
			if (i >= ai->num_hits)
				break;

			ship2 = player[next_player].ship_at[ai->hit[i].x][ai->hit[i].y];
			//Check if the new ship is in the queue
			if (ship == ship2)
				new = false;

			//Has the ship been sunk?
			if (player[next_player].ship[ship2].state == SHIP_SUNK)
			{
				--ai->num_hits;
				ai->hit[i] = ai->hit[ai->num_hits];
				//update ship origin
				if (i == 0)
				{
					ai->x = ai->hit[i].x;
					ai->y = ai->hit[i].y;
				}
			}
			else
			{
				++i;
			}
		}
		
		if (new)
		{
			//its a new ship add it to the queue
			ai->hit[ai->num_hits].x = ai->x;
			ai->hit[ai->num_hits].y = ai->y;
			//clear attemped directions
			size_t i;
			for (i = 0; i < 4; i++)
				ai->hit[ai->num_hits].dir[i] = false;
			//pick random direction
			ai->hit[ai->num_hits].curr_dir = (unsigned char)((double)4*rand()/(RAND_MAX+1.0));
			++ai->num_hits;
		}
	}
	//Miss, but only care if there is currently something in the hit queue
	else if (ai->num_hits > 0)
	{
		//Need to start looking in another direction.
		ai->hit[0].dir[ai->hit[0].curr_dir] = true;
		
		//Switch to opposite dir for now. If this has been tired it
		//wiil be corrected below
		ai->hit[0].curr_dir = (ai->hit[0].curr_dir + 2) % 4;

		//Move back to origin
		ai->x = ai->hit[0].x;
		ai->y = ai->hit[0].y;
	}

	//Determine next fireing locaiton
	while (ai->num_hits > 0 && !ai->chosen)
	{
		//pick a new direction if need be. 
		size_t count = 0;
		while (ai->hit[0].dir[ai->hit[0].curr_dir])
		{
			ai->hit[0].curr_dir = (ai->hit[0].curr_dir + 1) % 4;
			//start at first hit again
			ai->x = ai->hit[0].x;
			ai->y = ai->hit[0].y;

			//just in case
			if (++count > 4)
			{
				//DPRINT("Error: AI failed to sink ship\n");
				return;
			}
		}

		//move
		switch (ai->hit[0].curr_dir)
		{
			//Left
			case 3: 
				if (ai->x > 0) 
					--ai->x;
				else 
					ai->hit[0].dir[ai->hit[0].curr_dir] = true;
			break;	
			//Right
			case 1: 
				if (ai->x < GRID_DIM - 1) 
					++ai->x;
				else 
					ai->hit[0].dir[ai->hit[0].curr_dir] = true;
			break;	
			//Up
			case 0: 
				if (ai->y > 0) 
					--ai->y;
				else 
					ai->hit[0].dir[ai->hit[0].curr_dir] = true;
			break;	
			//Down
			case 2: 
				if (ai->y < GRID_DIM - 1) 
					++ai->y;
				else 
					ai->hit[0].dir[ai->hit[0].curr_dir] = true;
			break;	
			default:
				//DPRINT("Error: Invalid AI direction (%i)\n", 
				//		ai->hit[0].curr_dir);
				return;
			break;
		}

		if (player[next_player].grid[ai->x][ai->y] != GRID_EMPTY)
			ai->hit[0].dir[ai->hit[0].curr_dir] = true;
	
		//if the direction is still ok then its a valid move
		ai->chosen = !ai->hit[0].dir[ai->hit[0].curr_dir];
	}
}

static void
board_move(size_t axis, GLfloat to, GLfloat rate)
{
	board_motion.tend[axis] = to;
	float curr = battleship_view_get_translate(axis, 'g'); 
	board_motion.trate[axis] = (to > curr) ? rate : -rate;
	start_transition();
}

static void
board_rotate(size_t axis, GLfloat to, GLfloat rate)
{
	board_motion.rend[axis] = to;
	float curr = battleship_view_get_rotate(axis, 'g'); 
	board_motion.rrate[axis] = (to > curr) ? rate : -rate;
	start_transition();
}

static void
start_transition()
{
	if (game_state == GAME_TRANSITION)
		return;

	if (!do_trans_animation) cancel_animation = true;
	battleship_logic_start_timer(ANIMATION_SPEED, animate_transition);
}

static void
start_ai()
{
	game_state = GAME_AI_PLAYING;
	time_t speed = AI_SPEED;
	if (demo_mode)
		speed += DEMO_DELAY;
	battleship_logic_start_timer(speed, ai_play);
}

static void
start_fireing()
{
	game_state = GAME_FIREING;
	time_t speed = ANIMATION_SPEED + 80 - game_rocket.speed*20;
	battleship_logic_start_timer(speed, animate_fireing);
}

static void
start_message()
{
	game_state = GAME_MESSAGE;
	time_t speed = ANIMATION_SPEED;
	if (demo_mode)
		speed += DEMO_DELAY;
	battleship_logic_start_timer(speed, animate_message);
}

static void
autoplace_ships()
{
	size_t i;
	for (i = 0; i < NUM_SHIPS; i++)
	{
		do
		{
			player[curr_player].ship[i].x = (unsigned char) ((double) GRID_DIM*rand()/(RAND_MAX+1.0));
			player[curr_player].ship[i].y = (unsigned char) ((double) GRID_DIM*rand()/(RAND_MAX+1.0));
			player[curr_player].ship[i].down = (rand()/(RAND_MAX+1.0) > 0.5);
		} while (!try_place_ship(i));
	}

}
static bool
animate_transition()
{
	if (curr_menu || !do_trans_animation)
		cancel_animation = true;

	static enum battleship_gamestate trans_saved_state = GAME_PLACE_SHIPS;
	if (game_state != GAME_TRANSITION)
	{
		trans_saved_state = game_state;
		game_state = GAME_TRANSITION;
	}
	
	size_t count = 0;
	size_t i = 0;
	for (i = 0; i < 3; i++)
	{
		float curr = battleship_view_get_rotate(i, 'g');
		if (cancel_animation || 
		(board_motion.rrate[i] < 0.0 && curr <= board_motion.rend[i]) ||
		(board_motion.rrate[i] > 0.0 && curr >= board_motion.rend[i]) ||
		(board_motion.rrate[i] == 0.0)
		)
		{
			battleship_view_set_rotate(i, board_motion.rend[i], 'g'); 
			++count;
		}
		else
		{
			battleship_view_set_rotate(i, board_motion.rrate[i], 'G'); 
		}

		curr = battleship_view_get_translate(i, 'g');
		if (cancel_animation ||
		(board_motion.trate[i] < 0.0 && curr <= board_motion.tend[i]) ||
		(board_motion.trate[i] > 0.0 && curr >= board_motion.tend[i]) ||
		(board_motion.trate[i] == 0.0)
		)
		{
			battleship_view_set_translate(i, board_motion.tend[i], 'g'); 
			++count;
		}
		else
		{
			battleship_view_set_translate(i, board_motion.trate[i],'G'); 
		}
	}
	battleship_view_refresh();

	bool done = (count == 6);
	if (cancel_animation) cancel_animation = false;

	if (done) game_state = trans_saved_state;
	return !done;
}

static bool
animate_fireing()
{
	if (curr_menu || game_rocket.type == ROCKET_OFF)
		cancel_animation = true;
	
	battleship_view_refresh();
	bool done = (cancel_animation || !battleship_rocket_move());
	
	cancel_animation = false;
	if (done) next_state();
	return !done;
}

static bool
animate_message()
{
	if (curr_menu || !do_msg_animation)
		cancel_animation = true;

	//If the animation is canclled call back once more at full size
	if (cancel_animation && game_message.size < MAX_MSG_SIZE)
	{
		game_message.size = MAX_MSG_SIZE;
		battleship_view_refresh();
		return true;
	}
	//Grow the text
	else if (game_message.size < MAX_MSG_SIZE)
	{
		++game_message.size;
		battleship_view_refresh();
	}
	//Wait a bit
	else if (game_message.delay < MAX_MSG_DELAY)
	{
		++game_message.delay;
	}
	
	bool done = (cancel_animation || game_message.delay == MAX_MSG_DELAY);
	//stop sound if animation stoppe
	if (cancel_animation)
		battleship_sound_stop();
	cancel_animation = false;
	if (done) next_state();
	return !done;
}

