/* Matthew Phillips mephilli 20066500 CS488 A5
 *
 * Copyright (c) 2005 Matthew Phillips
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/** @file battleship_logic.h 
 *
 * Battleship logic.
 * 
 * These files are responsible for the majority of the game logic. (Some game
 * logic is handled by battleship_menu.c.)
 */
#ifndef BATTLEHIP_LOGIC_H
#define BATTLEHIP_LOGIC_H

#include <stdbool.h>
#include <time.h>

/** Constants used for mouse buttons */
enum battleship_mouse_button
{
	BATTLESHIP_MOUSE_LEFT = 0x01,
	BATTLESHIP_MOUSE_MIDDLE = 0x02,
	BATTLESHIP_MOUSE_RIGHT = 0x04
};

/** Constants used for keyboard modifiers */
enum battleship_input_mod
{
	BATTLESHIP_MOD_SHIFT = 0x01,
	BATTLESHIP_MOD_CRTL = 0x02,
	BATTLESHIP_MOD_ALT = 0x04,
	BATTLESHIP_MOD_AI = 0xff
};

/** This enum is used to index into the BATTLESHIP_KEYS. */
enum battleship_keynames
{
	BATTLESHIP_KEY_QUIT,
	BATTLESHIP_KEY_RESET,
	BATTLESHIP_KEY_LEFT,
	BATTLESHIP_KEY_RIGHT,
	BATTLESHIP_KEY_UP,
	BATTLESHIP_KEY_DOWN,
	BATTLESHIP_KEY_ROTATE,
	BATTLESHIP_KEY_ENTER,
	BATTLESHIP_KEY_ESC,
	BATTLESHIP_KEY_BACKSPACE
};

/** Defines keyboard values for game actions.
 *
 *  This array should be index using the enum battleship_keynames
 */
static const unsigned char BATTLESHIP_KEYS[] = 
{
	'Q',
	'R',
	1,
	2,
	3,
	4,
	5,
	13,
	27,
	8
};

/** Called to perform game initilization.
 * 
 *	This function initilizes game variables and must be
 *	called before drawing occurs. It inititizes all other
 *	parts of the game.
 * */
void 
battleship_logic_init();

/** Quits the game (implementation dependent)
 *
 *  This function should be implement in the main file corrispoding
 *  to a particualr gui frontend.
 *
 *  The function should cause the game to exit.
 */
extern void 
battleship_logic_quit();

/** Starts a new game.
 *
 *  This function does not need to be called directly.
 *  However it can be called directly to skip the intro menu.
 */
void
battleship_logic_new();

/** Starts a new one player game.
 *
 *  This function does not need to be called directly.
 *  However it can be called directly to skip the intro menu.
 */
void
battleship_logic_onep();

/** Starts a new two player game.
 *
 *  This function does not need to be called directly.
 *  However it can be called directly to skip the intro menu.
 */
void
battleship_logic_twop();

/** Starts a new game in demo mode.
 *
 *  This function does not need to be called directly. 
 *  However it can be called directly to skip the intro menu.
 */
void
battleship_logic_demo();

/** Resets the game.
 *
 * Regardless of the games current state it will be reset and return to
 * the opening menu
 */
void
battleship_logic_restart();

/** Handles mouse movement.
 *  
 *  This function should be called from the main file corrisponding to
 *  a particular gui frontend.
 *
 *  This function should be called whenever the mouse moves inside the
 *  window.
 *
 *  @param x The x coordinate of the mouse motion
 *  @param y The y coordinate of the mouse motion
 */
void 
battleship_logic_mouse_move(int x, int y);

/** Handles mouse button releases.
 *  
 *  This function should be called from the main file corrisponding to
 *  a particular gui frontend.
 *
 *  This function should be called whenever a mouse button is released.
 *
 *  @param button The button which was released. This value should corrispond to
 *  		      one of the values in the enum battleship_mouse_button.
 *
 *  @param mod What modifiers are were pressed. This value should corrispond to
 *             one of the values in the enum battleship_key_mod.
 *  @param x The x coordinate of the mouse event 
 *  @param y The y coordinate of the mouse event
 */
void 
battleship_logic_mouse_up(unsigned int button, unsigned int mod, int x, int y);

/** Handles mouse button presses.
 *  
 *  This function should be called from the main file corrisponding to
 *  a particular gui frontend.
 *
 *  This function should be called whenever a mouse button is pressed.
 *
 *  @param button The button which was pressed. This value should corrispond to
 *  		      one of the values in the enum battleship_mouse_button.
 *
 *  @param mod What modifiers are were pressed. This value should corrispond to
 *             one of the values in the enum battleship_key_mod.
 *  @param x The x coordinate of the mouse event 
 *  @param y The y coordinate of the mouse event
 */
void 
battleship_logic_mouse_down(unsigned int button,unsigned int mod, int x, int y);

/** Handles keypresses 
 *  
 *  This function should be called from the main file corrisponding to
 *  a particular gui frontend.
 *
 *  This function should be called whenever a key is pressed
 *
 *  @param c   The key that was pressed. This value should corrispond to
 *  			  a value in the BATTLESHIP_KEYS array. Which may be indexed
 *  			  with the enum battleship_keynames
 *
 *  @param mod What modifiers are were pressed. This value should corrispond to
 *             one of the values in the enum battleship_key_mod.
 */
void
battleship_logic_keypress(unsigned char c, unsigned int mod);

/** Start a timer (implementation dependent)
 *
 *  This function should be implement in the main file corrispoding
 *  to a particualr gui frontend.
 *
 *  The function should start a timer. The timer will call the 
 *  given function every delay ms until the function returns false.
 *
 *  @param delay How often to call the timer
 *  @param func The function to call
 */
extern void
battleship_logic_start_timer(time_t delay, bool (*func)(void));

/** Stops all timers (implementation dependent)
 *
 *  This function should be implement in the main file corrispoding
 *  to a particualr gui frontend.
 *
 *  The function should stop all current timers. As quickly as possible.
 *
 */
extern void
battleship_logic_stop_timers();

#endif /* BATTLEHIP_LOGIC_H */
