/* Matthew Phillips mephilli 20066500 CS488 A5
 *
 * Copyright (c) 2005 Matthew Phillips
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/** @file battleship_model.h 
 *
 *  These files are responsible for the data structures used by the game.
 */
#ifndef BATTLEHIP_MODEL_H
#define BATTLEHIP_MODEL_H

#include <stdio.h>
#include <stdbool.h>
#include <ctype.h>

#include "lsystem.h"
#include "psystem.h"
#include "battleship_rocket.h"

/*
#ifdef DEBUG
	// Debug output
    #define DPRINT(s, args...) fprintf(stderr, s, ##args)
#else
	// Debug output (Debuggin off)
    #define DPRINT(s, args...) if (false) {}; 
#endif
*/

/** The name of the Game */
#define BATTLESHIP_NAME "Battleship"
/** The current version */
#define BATTLESHIP_VERSION "v1.0"

/** The width/height of the playing grid */
#define GRID_DIM 10

/** Length of a carrier */
#define CARRIER_LEN 5
/** Length of a battleship */
#define BATTLESHIP_LEN 4
/** Length of a destroyer */
#define DESTROYER_LEN 3
/** Length of a sub */
#define SUB_LEN 3
/** Length of a pt boat */
#define PT_LEN 2

/** The number of ship types */
#define NUM_SHIPS 5
/** The length of the longest ship */
#define MAX_SHIP_LEN 5
/** The maximum number of possible misses */
#define MAX_MISSES GRID_DIM*GRID_DIM
/** The maxiumum number of possible hits */
#define MAX_HITS (CARRIER_LEN + BATTLESHIP_LEN + DESTROYER_LEN + SUB_LEN + PT_LEN)

/** Maximum length for a player name */
#define MAX_NAME_LEN 9 

/** The possible states the game may be in */
enum battleship_gamestate
{
	GAME_INIT,
	GAME_TRANSITION,
	GAME_PLACE_SHIPS,
	GAME_PLAYING,
	GAME_AI_PLAYING,
	GAME_FIREING,
	GAME_MESSAGE,
	GAME_OVER
};

/** How many AI types there are */
#define NUM_AI_TYPES 4
/** The possible types of AI */
enum battleship_aitype
{
	AI_HUMAN = 0,
	AI_EASY,
	AI_MEDIUM,
	AI_HARD
};

/** String representation of AI types 
 *	NOTE: Strings are padded with spaces to help menu centering
 * */
static const char *battleship_aitype_s[NUM_AI_TYPES] = 
{
	"Human    ",
	"Easy_AI  ",
	"Medium_AI",
	"Hard_AI  ",
};

/** How many test types there are */
#define NUM_TEST_TYPES 10
/** The possible types of tests*/
enum battleship_testtype
{
	TEST_NONE = 0,
	TEST_PRIMITIVE,
	TEST_CLOCK,
	TEST_MUG,
	TEST_PEG,
	TEST_SHIPS,
	TEST_FONT,
	TEST_FOG,
	TEST_LSYS,
	TEST_ROCKET,
};

/** String representation of test types */
static const char *battleship_testtype_s[NUM_TEST_TYPES] = 
{
	"None",
	"Primitive",
	"Clock",
	"Mug", 
	"Peg",
	"Ships",
	"Font",
	"Fog",
	"Lsystem",
	"Rocket",
};
/** Possible states for a ship */
enum battleship_shipstate
{
	SHIP_NOT_PLACED,
	SHIP_PLACING,
	SHIP_PLACED,
	SHIP_SUNK
};

/** Possible states for a grid block */
enum battleship_gridstate
{
	GRID_EMPTY,
	GRID_MISS,
	GRID_HIT
};

/** The types of ships */
enum battleship_shiptype
{
	CARRIER = 0,
	BATTLESHIP,
	DESTROYER,
	SUB,
	PT,
	NO_SHIP
};

/** String representation of Ship types */
static const char *battleship_shiptype_s[NUM_SHIPS] = 
{
	"Carrier",
	"Battleship",
	"Destroyer",
	"Sub",
	"PT Boat",
};

/** How many fog types there are */
#define NUM_FOG_TYPES 4
/** The possible types of fog */
enum battleship_fogtype
{
	FOG_OFF = 0,
	FOG_LIGHT,
	FOG_MEDIUM,
	FOG_HEAVY,
	FOG_REGEN
};
/** String representation of fog types */
static const char *battleship_fogtype_s[NUM_FOG_TYPES] = 
{
	"Off   ",
	"Light ",
	"Medium",
	"Heavy ",
};

/** The number of places shadows are draw */
#define NUM_SHADOWS 5
/** Enumeration to index into the do_shadows array to enable/disable different
 *  shadows
 */
enum battleship_shadow
{
	SHADOW_ALL = 0,
	SHADOW_TABLE,
	SHADOW_BACK_WALL,
	SHADOW_FRONT_WALL,
	SHADOW_FLOOR
};

/** How many rocket paths there are */
#define NUM_ROCKET_TYPES 3
/** The possible rocket paths */
enum battleship_rockettype
{
	ROCKET_OFF = 0,
	ROCKET_SIDE,
	ROCKET_TOP,
};
/** String representation of rocket paths */
static const char *battleship_rockettype_s[NUM_ROCKET_TYPES] = 
{
	"Off ",
	"Top ",
	"Side",
};
/** Minimun rocket size */
#define MIN_ROCKET_SIZE 1
/** Maximum rocket size */
#define MAX_ROCKET_SIZE 5 
/** Minimun rocket speed */
#define MIN_ROCKET_SPEED 1
/** Maximum rocket speed */
#define MAX_ROCKET_SPEED 5
/** Minimun rocket size */
#define MIN_FIRE_LENGTH 2
/** Maximum rocket size */
#define MAX_FIRE_LENGTH 6 
/** How many fire widths there are */
#define NUM_FIRE_WIDTHS 3
/** The possible rocket paths */
enum battleship_firewidth
{
	FIRE_SKINNY,
	FIRE_EQUAL,
	FIRE_WIDE,
};
/** String representation of fire widths */
static const char *battleship_firewidth_s[NUM_FIRE_WIDTHS] = 
{
	"Skinny",
	"Equal ",
	"Wide  ",
};

/** Represents a ship */
struct battleship_ship
{
	/** If the ship is pointing down (as opposed to left/right) */
	bool down;
	/** The x,y location of the ship (top/left x,y coordinate) */
	unsigned char x, y;
	/** How many times the ship has been hit */
	unsigned char num_hits;
	/** The length of the ship */
	unsigned char length;
	/** The ships current state */
	enum battleship_shipstate state;
};

/** Represents a player */
struct battleship_player
{
	/** The players name */
	char name[MAX_NAME_LEN + 1];
	/** The player's ships */
	struct battleship_ship ship[NUM_SHIPS]; 
	/** Stores where, on this players grid the opponent has fired
	 *  and if it was a hit or a miss
	 */
	enum battleship_gridstate grid[GRID_DIM][GRID_DIM];
	/** Stores what ship, if any is at each grid location */
	enum battleship_shiptype ship_at[GRID_DIM][GRID_DIM];
	/** Number of hits this player has recieved */
	unsigned char num_hits;
	/** Number of missses this player has recieved */
	unsigned char num_misses;
	/** Number of ships of this player's that have been sunk */
	unsigned char num_sunk;
	/** The AI type for this player */
	enum battleship_aitype ai;
	/** Set to true to have the players ships placed aoutmatically */
	bool auto_place;
	/** Set to true if the players ships should be obsucred */
	bool fog;
	/** The x coordinate of the current selected grid location */
	unsigned char sel_x;
	/** The y coordinate of the current selected grid location */
	unsigned char sel_y;
};

/** Used to represent hit/miss/sunk messages */
struct battleship_message
{
	/** String representation of the ship that was hit */
	char *ship;
	/** The type of hit hit/miss/none, none is used for game over */
	enum battleship_gridstate type;
	/** This value is used by the animation to grow the text */
	float size;
	/** This value is used by the animation to allow the text to persist for
	 *  a period once it has reached its maximum size */
	unsigned char delay;
};

/** Used for storing data for the name selector */
struct battleship_name_selector
{
	/** True when the selector is enabled (it will be drawn */	
	bool enabled;
	/** The x location of the selected character */
	int x;
	/** The y location of the selected character */
	int y;
	/** The name that is being selected */
	char *name;
};

/** Minum length for lsystem */
#define MIN_LSYSTEM_LENGTH 1
/** Maximum length for lsystem */
#define MAX_LSYSTEM_LENGTH 10
/** Structer defining parametrs for the lsystem used for drawing on the walls */
struct battleship_lsystem
{
	/** If lsystem is enabled */
	bool enabled;
	/** Which to use */
	unsigned char type;
	/** The starting line length */
	unsigned char length;
	/** String represenation of length */
	char slength[3];
};

/** Structure for holding parameters related to rocket */
struct battleship_rocket
{
	/** The path the rocket should use */
	enum battleship_rockettype type;
	/** True if the path should be drawn */
	bool show_path;
	/** The size of the rocket */
	unsigned char size;
	/** String representation of size */
	char ssize[2];
	/** If the rocket should have fire */
	bool draw_fire;
	/** The length of rocket fire */
	unsigned char len;
	/** String representation of length */
	char slen[2];
	/** The radius of rocket fire */
	enum battleship_firewidth width;
	/** The speed of the rocket animation */
	unsigned char speed;
	/** String reresetation of speed */
	char sspeed[2];
	/** The colour of the fire */
	enum psystem_colour colour;
};

/** Keeps track of the game state */
enum battleship_gamestate game_state;
/** The players */
struct battleship_player player[2];
/** The current player */
unsigned char curr_player;
/** The current game message */
struct battleship_message game_message;
/** Name selector */
struct battleship_name_selector name_selector;
/** Lsystem struct */
struct battleship_lsystem game_lsys;
/** Rocket struct */
struct battleship_rocket game_rocket;

/** True when demo mode is active */
bool demo_mode;
/** The type of fog to use */
enum battleship_fogtype do_fog;
/** True if game board movement shoudl be animated */
bool do_trans_animation;
/** True if AI movement should be animated */
bool do_ai_animation;
/** True if message animation should take place */
bool do_msg_animation;
/** What test to run */
enum battleship_testtype do_test;
/** True if textures should be drawn */
bool do_textures;
/** [0] is true if shadows are enable. The rest of the indecies are
 *  for enabling different shadows, see enum.*/
bool do_shadows[NUM_SHADOWS];

/** Inititlizes the game data structures.
 *
 *  Called by battleship_logic_init(). Does not need to be
 *  called directly.
 */
void
battleship_model_init();

#endif /* BATTLEHIP_MODEL_H */
