/* Matthew Phillips mephilli 20066500 CS488 A5
 *
 * Copyright (c) 2005 Matthew Phillips
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/** @file glprimitive.c 
 *
 *  Draw curved primitives with splines.
 */
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <math.h>
#include <time.h>

#include "glprimitive.h"

/** The number of control points */
static const size_t NUM_POINTS = 3;
/** The amount of detail for splines */
size_t detail = 5;
/** Additional wait added to points */
#define WEIGHT 1.265
/** This is 1/sqrt(2) = cos(pi/4) = sin(pi/4) */
static const GLdouble TRIG = 0.70710678118654746172 * WEIGHT; 
/** This TRIG*TRIG. Currently 1/2 */
static const GLdouble TRIG2 = 0.5 * WEIGHT * WEIGHT; 

/** Draws a sphere.
 *
 *  glprimitive_sphere and glprimitive_half_sphere are just wrappers for this
 *  function.
 *
 *  @param radius The radius of the sphere
 *  @param half Set to true to only draw half of the sphere
 */  
inline static void sphere(GLdouble radius, bool half);

/** Draws a torus.
 *
 *  glprimitive_torus and glprimitive_half_torus are just wrappers for this
 *  function.
 *
 *  @param r1 The major radius of the torus
 *  @param r2 The minor radius of the torus
 *  @param half1 Set to true to only draw torus on positive side of z = 0
 *  @param half2 Set to true to only draw torus on positive side of x = 0
 */  
inline static void torus(GLdouble r1, GLdouble r2, bool half1, bool half2);

/** Draws a cone/cylinder.
 *
 *  glprimitive_cylinder and glprimitive_cone are just wrappers for this
 *  function.
 *
 *  The cone is centered so that half its length is on either side of the
 *  origin (with its opening point along z). The minor radius is behind the 
 *  origin and the major in front. 
 * 
 *  @param r1 The major radius of the cone
 *  @param r2 The minor radius of the minor
 *  @param len The length of the cone.
 */  
inline static void cone(GLdouble r1, GLdouble r2, GLdouble len);

void
glprimitive_set_detail(size_t d)
{
	detail = d;
}

size_t
glprimitive_get_detail()
{
	return detail;
}

void
glprimitive_box(GLdouble x, GLdouble y, GLdouble z, 
				 GLdouble w, GLdouble h, GLdouble d)
{
	GLdouble s = 1.0/(h + w);
	GLdouble t = 1.0/(d + h);
	
	glBegin(GL_QUADS);
		//Back
		glNormal3f(0, 0, -1);
		glTexCoord2f(1, 1); glVertex3f(x, y + h, z);
		glTexCoord2f(h*s, 1); glVertex3f(x + w, y + h, z);
		glTexCoord2f(h*s, d*t); glVertex3f(x + w, y, z);
		glTexCoord2f(1, d*t); glVertex3f(x, y, z);
		//Front
		glNormal3f(0, 0, 1);
		glTexCoord2f(h*s, 1); glVertex3f(x, y, z + d);
		glTexCoord2f(1, 1); glVertex3f(x + w, y, z + d);
		glTexCoord2f(1, d*t); glVertex3f(x + w, y + h, z + d);
		glTexCoord2f(h*s, d*t); glVertex3f(x, y + h, z + d);
		//Left
		glNormal3f(-1, 0, 0);
		glTexCoord2f(0, 0); glVertex3f(x, y, z);
		glTexCoord2f(0, d*t); glVertex3f(x, y, z + d);
		glTexCoord2f(h*s, d*t); glVertex3f(x, y + h, z + d);
		glTexCoord2f(h*s, 0); glVertex3f(x, y + h, z);
		//Right
		glNormal3f(1, 0, 0);
		glTexCoord2f(h*s, 0); glVertex3f(x + w, y, z + d);
		glTexCoord2f(h*s, d*t); glVertex3f(x + w, y, z);
		glTexCoord2f(0, d*t); glVertex3f(x + w, y + h, z);
		glTexCoord2f(0, 0); glVertex3f(x + w, y + h, z + d);
		//Top
		glNormal3f(0, 1, 0);
		glTexCoord2f(h*s, 0); glVertex3f(x, y + h, z); 
		glTexCoord2f(h*s, d*t); glVertex3f(x, y + h, z + d);
		glTexCoord2f(1, d*t); glVertex3f(x + w, y + h, z +d);
		glTexCoord2f(1, 0); glVertex3f(x + w, y + h, z);
		//Bottom
		glNormal3f(0, -1, 0);
		glTexCoord2f(h*s, d*t); glVertex3f(x + w, y, z);
		glTexCoord2f(h*s, 0); glVertex3f(x + w, y, z + d);
		glTexCoord2f(1, 0); glVertex3f(x, y, z + d);
		glTexCoord2f(1, d*t); glVertex3f(x, y, z);
	glEnd();
}

void
glprimitive_disk(GLdouble radius)
{
	size_t D_NUM_POINTS = detail*4;
	
	size_t i;
	glBegin(GL_TRIANGLE_FAN);
		glNormal3f(0.0, 0.0, 1.0);
		glTexCoord2d(0.5,0.5);
		glVertex3f(0.0, 0.0, 0.0);
		for (i = 0; i <= D_NUM_POINTS; i++)
		{
			glTexCoord2d(0.5 + 0.5*cos(2.0*i*M_PI/D_NUM_POINTS), 
						 0.5 + 0.5*sin(2.0*i*M_PI/D_NUM_POINTS));
			glVertex3d(
						radius*cos(2.0*i*M_PI/D_NUM_POINTS),
						radius*sin(2.0*i*M_PI/D_NUM_POINTS),
						0.0
					);
		}
	glEnd();

/* 
	//This draws a disc with splines. But it does not work with a display list 
	//for reasons I have been unable to determine. So I used the above method
	//which is likely faster anyway

	GLdouble ctrlpoints[][3] =
	{
		{radius, 0.0, 0.0},
		{radius*TRIG, radius*TRIG, 0.0},
		{0.0, radius, 0.0}
	};

	glEnable(GL_MAP1_VERTEX_3);
	glMap1d(GL_MAP1_VERTEX_3, 0.0, 1.0, 3, NUM_POINTS, &ctrlpoints[0][0]);

	size_t r, i;
	glPushMatrix();
	for (r = 0; r < 4; r++)
	{
		glBegin(GL_TRIANGLE_FAN);
		glNormal3f(0.0, 0.0, 1.0);
		glVertex3f(0.0, 0.0, 0.0);
		for (i = 0; i <= detail; i++)
		{
			glEvalCoord1d((GLdouble) i/detail);
		}
		glEnd();
		glRotatef(90.0, 0.0, 0.0, 1.0);
	}
	glPopMatrix();

	glDisable(GL_MAP1_VERTEX_3);
*/
}

static void
cone(GLdouble r1, GLdouble r2, GLdouble len)
{
	GLfloat ctrlpoints[2][3][3] =
	{
		{
			{r1, 0.0, 0.0},
			{r1*TRIG, r1*TRIG, 0.0},
			{0.0, r1, 0.0}
		},
		{
			{r2, 0.0, len},
			{r2*TRIG, r2*TRIG, len},
			{0.0, r2, len}
		},
	};

	glPushAttrib(GL_ENABLE_BIT);
	glEnable(GL_MAP2_VERTEX_3);
	glEnable(GL_AUTO_NORMAL);
	glMap2f(GL_MAP2_VERTEX_3, 0, 1, 3, NUM_POINTS, 
							  0, 1, NUM_POINTS*3, 2, 
							  &ctrlpoints[0][0][0]);
	glMapGrid2f(detail, 0.0, 1.0, 1, 0.0, 1.0);

	size_t r;
	glPushMatrix();
	for (r = 0; r < 4; r++)
	{
		glEvalMesh2(GL_FILL, 0.0, detail, 0.0, 1);
		glRotatef(90.0, 0.0, 0.0, 1.0);
	}
	glPopMatrix();

	glPopAttrib();
}

void
glprimitive_cone(GLdouble r1, GLdouble r2, GLdouble len)
{
	cone(r1, r2, len);
}

void
glprimitive_cylinder(GLdouble radius, GLdouble len)
{
	cone(radius, radius, len);
}

inline static void
sphere(GLdouble radius, bool half)
{
	GLfloat ctrlpoints[3][3][3] =
	{
		{
			{radius, 0.0, 0.0},
			{radius*TRIG,radius*TRIG,0.0},
			{0.0, radius, 0.0}
		},
		{
			{radius*TRIG, 0.0, radius*TRIG},
			{
				radius*TRIG2, 
				radius*TRIG, 
				radius*TRIG2
			},
			{0.0, radius, 0.0}
		},
		{
			{0.0, 0.0, radius},
			{0.0, radius*TRIG, radius*TRIG},
			{0.0, radius, 0.0}
		},
	};

	glPushAttrib(GL_ENABLE_BIT);
	glEnable(GL_MAP2_VERTEX_3);
	glEnable(GL_MAP2_NORMAL);
	glEnable(GL_NORMALIZE);
	glMap2f(GL_MAP2_VERTEX_3, 0, 1, 3, NUM_POINTS, 
							  0, 1, NUM_POINTS*3, 3, 
							  &ctrlpoints[0][0][0]);
	glMap2f(GL_MAP2_NORMAL, 0, 1, 3, NUM_POINTS, 
							  0, 1, NUM_POINTS*3, 3, 
							  &ctrlpoints[0][0][0]);
	glMapGrid2f(detail, 0.0, 1.0, detail, 0.0, 1.0);

	size_t i, r;
	glPushMatrix();
	i = (half) ? 1 : 2;
	for ( ; i > 0; i--)
	{
		for (r = 0; r < 4; r++)
		{
			glEvalMesh2(GL_FILL, 0.0, detail, 0.0, detail);
			glRotatef(90.0, 0.0, 0.0, 1.0);
		}
		glRotatef(180, 0, 1, 0);
	}
	glPopMatrix();
	
	glPopAttrib();
}

void
glprimitive_sphere(GLdouble radius)
{
	sphere(radius, false);
}

void
glprimitive_half_sphere(GLdouble radius)
{
	sphere(radius, true);
}

inline static void
torus(GLdouble r1, GLdouble r2, bool half1, bool half2)
{
	r1 = r1 + r2;
	GLfloat ctrlpoints[3][3][3] =
	{
		{
			{r1 + r2, 0.0, 0.0},
			{r1 + r2*TRIG, r2*TRIG,0.0},
			{r1, r2, 0.0}
		},
		{
			{r1*TRIG + r2*TRIG, 0.0, r1*TRIG + r2*TRIG},
			{
				r1*TRIG + r2*TRIG2, 
				r2*TRIG, 
				r1*TRIG + r2*TRIG2
			},
			{r1*TRIG, r2, r1*TRIG}
		},
		{
			{0, 0, r1 + r2},
			{0, r2*TRIG, r1 + r2*TRIG},
			{0, r2, r1}
		},
	};
	
	glPushAttrib(GL_ENABLE_BIT);
	glEnable(GL_MAP2_VERTEX_3);
	glEnable(GL_AUTO_NORMAL);
//	glEnable(GL_MAP2_NORMAL);
//	glEnable(GL_NORMALIZE);
	glMap2f(GL_MAP2_VERTEX_3, 0, 1, 3, NUM_POINTS, 
							  0, 1, NUM_POINTS*3, 3, 
							  &ctrlpoints[0][0][0]);
//	glMap2f(GL_MAP2_NORMAL, 0, 1, 3, NUM_POINTS, 
//							  0, 1, NUM_POINTS*3, 3, 
//							  &ctrlpoints[0][0][0]);
	glMapGrid2f(detail, 0.0, 1.0, detail, 0.0, 1.0);

	size_t i, r;
	//Outter portion
	glPushMatrix();
	glRotatef(90, 1.0, 0.0, 0.0);
	for (i = 0; i < 2; i++)
	{
		for (r = 0; r < 4; r++)
		{
			if ((!half1 || r < 2))
				glEvalMesh2(GL_FILL, 0.0, detail, 0.0, detail);
			glRotatef(90.0, 0.0, 1.0, 0.0);
		}
		glRotatef(180, 1, 0, 0);
		if (half2) break;
	}
	glPopMatrix();

	r2 *= -1;
	GLfloat ctrlpoints2[3][3][3] =
	{
		{
			{r1 + r2, 0.0, 0.0},
			{r1 + r2*TRIG, r2*TRIG,0.0},
			{r1, r2, 0.0}
		},
		{
			{r1*TRIG + r2*TRIG, 0.0, r1*TRIG + r2*TRIG},
			{
				r1*TRIG + r2*TRIG2, 
				r2*TRIG, 
				r1*TRIG + r2*TRIG2
			},
			{r1*TRIG, r2, r1*TRIG}
		},
		{
			{0, 0, r1 + r2},
			{0, r2*TRIG, r1 + r2*TRIG},
			{0, r2, r1}
		},
	};
	
	glMap2f(GL_MAP2_VERTEX_3, 0, 1, 3, NUM_POINTS, 
							  0, 1, NUM_POINTS*3, 3, 
							  &ctrlpoints2[0][0][0]);
//	glMap2f(GL_MAP2_NORMAL, 0, 1, 3, NUM_POINTS, 
//							  0, 1, NUM_POINTS*3, 3, 
//							  &ctrlpoints2[0][0][0]);
	glMapGrid2f(detail, 0.0, 1.0, detail, 0.0, 1.0);
	
	//Inner protion
	glPushMatrix();
	glRotatef(90, 1.0, 0.0, 0.0);
	if (half2) glRotatef(180, 1.0, 0.0, 0.0);
	for (i = 0; i < 2; i++)
	{
		for (r = 0; r < 4; r++)
		{
			if ((!half1 || r < 2))
				glEvalMesh2(GL_FILL, 0.0, detail, 0.0, detail);
			glRotatef(90.0, 0.0, 1.0, 0.0);
		}
		glRotatef(180, 1, 0, 0);
		if (half2) break;
	}
	glPopMatrix();

	glPopAttrib();
}

void
glprimitive_torus(GLdouble r1, GLdouble r2)
{
	torus(r1, r2, false, false);
}

void
glprimitive_half_torus1(GLdouble r1, GLdouble r2)
{
	torus(r1, r2, true, false);
}

void
glprimitive_half_torus2(GLdouble r1, GLdouble r2)
{
	torus(r1, r2, false, true);
}

void
glprimitive_test(GLdouble r, size_t d)
{
	static const float diff_r[] = {1.0, 0.0, 0.0, 1.0};
	static const float spec_r[] = {0.1, 0.1, 0.1, 1.0};
	static const float shinny_r = 1;
	static GLuint list = 0;
	if (list == 0)
	{
		list = glGenLists(1);
		size_t oldd = glprimitive_get_detail();
		glprimitive_set_detail(d);
		glNewList(list, GL_COMPILE);
			glMaterialfv(GL_FRONT, GL_DIFFUSE, diff_r);
		    glMaterialfv(GL_FRONT, GL_SPECULAR, spec_r);
			glMaterialf(GL_FRONT, GL_SHININESS, shinny_r);
			glprimitive_torus(r/2.0, r/2.0);
			glTranslatef(-2*r, 2*r, 0);
			glprimitive_cone(2.0, 4, 4);
			glTranslatef(4*r, 0, 0);
			glprimitive_cylinder(4, 4);
			glTranslatef(0, -4*r, 0);
			glprimitive_disk(4);
			glTranslatef(-4*r, 0, 0);
			glprimitive_sphere(4);
		glEndList();	
		glprimitive_set_detail(oldd);
	}
	glCallList(list);
}

void
glprimitive_clock(GLdouble size, size_t d)
{
	static GLdouble last_size = 0;
	static size_t last_d = 0;

	static const GLfloat diff_y[] = {1.0, 1.0, 0.0, 1.0};
	static const GLfloat spec_y[] = {0.5, 0.5, 0.5, 1.0};
	static const GLfloat shinny_y = 15.0;
	
	static const GLfloat diff_face[] = {0.7, 0.7, 0.7, 1.0};
	static const GLfloat spec_face[] = {0.1, 0.1, 0.1, 1.0};
	static const GLfloat shinny_face = 1;
	
	/*	
	static const float diff_glass[] = {0.5, 0.5, 0.5, 0.6};
	static const float spec_glass[] = {0.1, 0.1, 0.1, 0.6};
	static const float shinny_glass = 1.0;
	*/
	
	static const GLfloat diff_tick[] = {0.0, 0.0, 0.0, 1.0};
	static const GLfloat spec_tick[] = {0.0, 0.0, 0.0, 1.0};
	static const GLfloat shinny_tick = 1.0;
	
	size_t i;
	GLdouble r1 = (size-size/16.0);
	GLdouble r2 = (size-size/8.0);
	static GLuint list = 0;
	if (last_size != size || last_d != d)
	{
		if (list == 0)
			list = glGenLists(1);
		size_t oldd = glprimitive_get_detail();
		glprimitive_set_detail(d);
		glNewList(list, GL_COMPILE);
			//The front face
			glMaterialfv(GL_FRONT, GL_DIFFUSE, diff_face);
		    glMaterialfv(GL_FRONT, GL_SPECULAR, spec_face);
			glMaterialf(GL_FRONT, GL_SHININESS, shinny_face);
			glprimitive_disk(size*0.98);
			//The ticks
			 //I don't like how this is done. I should make it pretty
			glMaterialfv(GL_FRONT, GL_DIFFUSE, diff_tick);
			glMaterialfv(GL_FRONT, GL_SPECULAR, spec_tick);
			glMaterialf(GL_FRONT, GL_SHININESS, shinny_tick);
			glLineWidth(size/4.0);
			glBegin(GL_LINES);
			glNormal3f(0.0, 0.0, 1.0);
			for (i = 0; i < 60; i++)
			{
				GLdouble up = sin(2*i*M_PI/60);
				GLdouble over = cos(2*i*M_PI/60);
				if (i % 5 != 0)
				{
					glVertex3d(r1*over, r1*up, 0.1);
					glVertex3d(r2*over, r2*up, 0.1);
				}
				else
				{
					glVertex3d(r1*over, r1*up, 0.1);
					glVertex3d((r2-size/16.0)*over, (r2-size/16.0)*up, 0.1);
				}
			}
			glEnd();
			glLineWidth(size/2.0);
			r2 -= size/16.0;
			glBegin(GL_LINES);
			for (i = 0; i < 4; i++)
			{
				GLdouble up = sin(2*i*M_PI/4);
				GLdouble over = cos(2*i*M_PI/4);
				glVertex3d(r1*over, r1*up, 0.1);
				glVertex3d(r2*over, r2*up, 0.1);
			}
			glEnd();
			//Circle in center
			glTranslatef(0.0, 0.0, 0.3);
			glprimitive_disk(size/16.0);
			glTranslatef(0.0, 0.0, -0.3);
			//The glass
			/*
			glPushMatrix();
				glTranslatef(0, 0, size/4.0-0.1);
				glMaterialfv(GL_FRONT, GL_DIFFUSE, diff_glass);
			    glMaterialfv(GL_FRONT, GL_SPECULAR, spec_glass);
				glMaterialf(GL_FRONT, GL_SHININESS, shinny_glass);
				glprimitive_disk(size*0.98);
			glPopMatrix();	
			*/
			//The border
			glPushMatrix();
				glMaterialfv(GL_FRONT, GL_DIFFUSE, diff_y);
			    glMaterialfv(GL_FRONT, GL_SPECULAR, spec_y);
				glMaterialf(GL_FRONT, GL_SHININESS, shinny_y);
				glTranslatef(0, 0, -size/4.0);
				glprimitive_cylinder(size, size/2.0);
			glPopMatrix();
			//The back
			glPushMatrix();
				glTranslatef(0, 0, -size/4.0);
				glRotatef(180, 0.0, 1.0, 0.0);
				glprimitive_disk(size);
			glPopMatrix();	
			//The legs
			glPushMatrix();
			for (i = 0; i < 2; i++)
			{
				glPushMatrix();
					glTranslatef(-size - size/32.0,-size - size/16.0,-size/4.0);
					glRotatef(-45, 0.0, 0.0, 1.0);
					  //Bottom
					glprimitive_cylinder(size/16.0, size/2.0);
					  //Cap the bottom
					glPushMatrix();
						glTranslatef(0.0, 0.0, size/2.0);
						glprimitive_disk(size/16.0);
						glTranslatef(0.0, 0.0, -size/2.0);
						glRotatef(180, 0.0, 1.0, 0.0);
						glprimitive_disk(size/16.0);
					glPopMatrix();
					  //The left and right
					glRotatef(90, 1, 0, 0);
					glTranslatef(0, size/16.0, -size/2.0);
					glprimitive_cylinder(size/16.0, size/2.0);
					glTranslatef(0.0, size/2.0 - size/8.0, 0.0);
					glprimitive_cylinder(size/16.0, size/2.0);
				glPopMatrix();
				glRotatef(180, 0.0, 1.0, 0.0); 
			}
			glPopMatrix();
			//The bells
			glPushMatrix();
			for (i = 0; i < 2; i++)
			{
				glPushMatrix();
					glTranslatef(-size/2.0 - size/4.0 - size/16.0, 
								 size - size/4.0, 0);
					glRotatef(-45, 0.0, 0.0, 1.0);
					glRotatef(90, 0.0, 1.0, 0.0);
					  //Stick
					glprimitive_cylinder(size/16.0, size/8.0);
					  //Bell
					glRotatef(180, 0.0, 1.0, 0.0);
					glprimitive_half_sphere(size/4.0);
				glPopMatrix();
				glRotatef(180, 0.0, 1.0, 0.0); 
			}
			glPopMatrix();
			//The stick
			glPushMatrix();
				glTranslatef(0.0, size + size/8.0, 0.0);
				glRotatef(90, 1.0, 0.0, 0.0);
				glprimitive_cylinder(size/16.0, size/8.0);
				glTranslatef(0.0, 0.0, -size/8.0 + size/16.0);
				glprimitive_sphere(size/8.0);
			glPopMatrix();	
		glEndList();	
		glprimitive_set_detail(oldd);

		last_size = size;
		last_d = d;
	}
	glCallList(list);
	
	glMaterialfv(GL_FRONT, GL_DIFFUSE, diff_tick);
	glMaterialfv(GL_FRONT, GL_SPECULAR, spec_tick);
	glMaterialf(GL_FRONT, GL_SHININESS, shinny_tick);

	time_t t = time(0);
	struct tm *time_struct = localtime(&t);

	//Sec hand
	/*
	int seconds = time_struct->tm_sec;
	GLdouble up = sin(2 * (GLdouble) seconds * M_PI/60 + M_PI/2.0);
	GLdouble over = -cos(2 * (GLdouble) seconds * M_PI/60 + M_PI/2.0);
	glLineWidth(size/2.0);
	glBegin(GL_LINES);
		glVertex3d((r2-size/16.0)*over, (r2-size/16.0)*up, 0.2);
		glVertex3d(0, 0, 0.2);
	glEnd();
	*/
	//Min hand
	int min = time_struct->tm_min;
	GLdouble up = sin(2 * (GLdouble) min * M_PI/60 + M_PI/2.0);
	GLdouble over = -cos(2 * (GLdouble) min * M_PI/60 + M_PI/2.0);
	glLineWidth(size);
	glBegin(GL_LINES);
		glVertex3d(r2*over, r2*up, 0.2);
		glVertex3d(0, 0, 0.2);
	glEnd();
	//Hour hand
	int hour = time_struct->tm_hour*5 + time_struct->tm_min/12;
	up = sin(2 * (GLdouble) hour * M_PI/60 + M_PI/2.0);
	over = -cos(2 * (GLdouble) hour * M_PI/60 + M_PI/2.0);
	r2 -= size/8.0 + size/16.0;
	glLineWidth(2*size);
	glBegin(GL_LINES);
		glVertex3d(r2*over, r2*up, 0.2);
		glVertex3d(0, 0, 0.2);
	glEnd();
}

void 
glprimitive_mug(GLdouble size, size_t d)
{
	static GLdouble last_size = 0;
	static size_t last_d = 0;

	static const GLfloat diff_w[] = {0.7, 0.7, 0.7, 1.0};
	static const GLfloat spec_w[] = {0.4, 0.4, 0.4, 1.0};
	static const GLfloat shinny_w = 15.0;
	
//	static const GLfloat diff_b[] = {0.4, 0.2, 0.0, 1.0};
//	static const GLfloat spec_b[] = {0.1, 0.1, 0.1, 1.0};
//	static const GLfloat shinny_b = 1.0;

	static GLuint list = 0;
	if (last_size != size || last_d != d)
	{
		if (list == 0)
			list = glGenLists(1);
		size_t oldd = glprimitive_get_detail();
		glprimitive_set_detail(d);
		glNewList(list, GL_COMPILE);
		glPushMatrix();
			glMaterialfv(GL_FRONT, GL_DIFFUSE, diff_w);
			glMaterialfv(GL_FRONT, GL_SPECULAR, spec_w);
			glMaterialf(GL_FRONT, GL_SHININESS, shinny_w);
			//The handel
			glPushMatrix();
				glTranslatef(size - size/32.0, size/8.0, 0.0); 
				glprimitive_half_torus1(size/2.0, size/8.0);
			glPopMatrix();
			//Outside
			glTranslatef(0, -size, 0);
			glRotatef(-90, 1.0, 0.0, 0.0);
			glprimitive_cylinder(size, 2*size);
			//Bottom
			glRotatef(180, 1.0, 0.0, 0.0);
			glprimitive_disk(size*0.99);
			//Rim
			glRotatef(180, 1.0, 0.0, 0.0);
			glTranslatef(0.0, 0.0, 2*size);
			glprimitive_half_torus2(size - size/4.0, size/8.0);
			//Coffie
//			glMaterialfv(GL_FRONT, GL_DIFFUSE, diff_b);
//			glMaterialfv(GL_FRONT, GL_SPECULAR, spec_b);
//			glMaterialf(GL_FRONT, GL_SHININESS, shinny_b);
//			glprimitive_disk(size - size/4.0);
		glPopMatrix();
		glEndList();	
		glprimitive_set_detail(oldd);

		last_size = size;
		last_d = d;
	}
	glCallList(list);
}
