/* Matthew Phillips mephilli 20066500 CS488 A5
 *
 * Copyright (c) 2005 Matthew Phillips
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/** @file lsystem.c 
 *
 *  A lsystem generator  
 */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <math.h>
#include <string.h>

#include "lsystem.h"

/*
#ifndef DPRINT
	#ifdef DEBUG
		// Debug output
	    #define DPRINT(s, args...) fprintf(stderr, s, ##args)
	#else
		// Debug output (Debuggin off)
	    #define DPRINT(s, args...) if (false) {}; 
	#endif
#endif
*/

/** This structure holds a lot more information than the normal lsystem
 *  struct, It keeps track of the lsystem state through recursive calls 
 */
struct lsystem_state
{
	/** The width of the data area */
	size_t w;
	/** The height of the data area */
	size_t h;
	/** Current x location */
	size_t x;
	/** Current y location */
	size_t y;

	/** Current recursive depth */
	unsigned int depth;
	/** Current angle */
	float angle;
	/** Current length */
	float len;
	/** The current string  */
	char *str;

	/** The lsystem */
	struct lsystem *lsys;

	/** How to colour the lsystem */
	enum lsystem_colour_mode colour_mode;
	/** The current colour */
	unsigned char colour[3];
	/** A list of colours */
	unsigned char *colour_list;
	/** How many colours are in the list of colours */
	int num_colours;
	/** The index of the current colour in the list */
	int ci;

	/** The data area (where pixles are written to) */
	char *data;
};

/** Finds the x and y component of the current line slope 
 *  @param state The lsystem
 *  @param dx The location to return the x component of the slope
 *  @param dy The location to return the y component of the slope
 */
static inline void find_slope(struct lsystem_state *state, int *dx, int *dy)
{
	(*dx) = (int) rint(state->len*cos(state->angle*M_PI/180.0));
	(*dy) = (int) rint(state->len*sin(state->angle*M_PI/180.0));
}

/** Fills in a pixel 
 *  @param state The lsystem
 */
static inline void draw_pixel(struct lsystem_state *state)
{
	if (state->x < state->w && state->y < state->h)
	{
		(state->data)[(state->y*state->w + state->x)*3] = state->colour[0];
		(state->data)[(state->y*state->w + state->x)*3 + 1] = state->colour[1];
		(state->data)[(state->y*state->w + state->x)*3 + 2] = state->colour[2];
	}
}

/** Draws a line
 *  @param state The lsystem
 */
static void draw_line(struct lsystem_state *state)
{
	int dx, dy;
	find_slope(state, &dx, &dy);

	//where the line is going
	size_t x1 = state->x + dx;
	size_t y1 = state->y + dy;

	size_t *x, *y;
	//slope is greater than 1 move on y instead of x
	if (abs(dy) > abs(dx))
	{
		x = &state->y;
		y = &state->x;

		size_t temp = x1;
		x1 = y1;
		y1 = temp;

		int temp2 = dx;
		dx = dy;
		dy = temp2;
	}
	else
	{
		x = &state->x;
		y = &state->y;
	}

	//deal with negative slopes
	int xstep;
	int ystep;
	if (dx > 0) xstep = 1; else xstep = -1;
	if (dy > 0) ystep = 1; else ystep = -1;
	dx = abs(dx);
	dy = abs(dy);

	//keep track of error
	int e = 0;
	int de = dy;

	if (state->colour_mode == LSYSTEM_COLOUR_LINE)
	{
		state->ci = (state->ci + 1) % state->num_colours;
		state->colour[0] = state->colour_list[state->ci * 3];
		state->colour[1] = state->colour_list[state->ci * 3 + 1];
		state->colour[2] = state->colour_list[state->ci * 3 + 2];
	}

	while ((*x) != x1)
	{
		draw_pixel(state);
		(*x) += xstep;

		e += de;
		if (e << 1 >= dx)
		{
			(*y) += ystep;
			e -= dx;
		}
	}
	draw_pixel(state);
}

/** Draws an lsytem, called recursivly 
 *  @param state The lsystem
 */
static void lsystem_draw_r(struct lsystem_state *state)
{
	if (state->colour_mode == LSYSTEM_COLOUR_DEPTH)
	{
		state->ci = state->depth*(state->num_colours/((double)state->lsys->depth + 1.0));
		state->colour[0] = state->colour_list[state->ci * 3 + 0];
		state->colour[1] = state->colour_list[state->ci * 3 + 1];
		state->colour[2] = state->colour_list[state->ci * 3 + 2];
	}
	
	char *c = state->str;
	
	bool done = false;
	int negate = 1;
	while (*c && !done)
	{
		bool angle = false;
		char *search_i = NULL;
		
		//If haven't reach the maximum recursive depth
		if (state->depth < state->lsys->depth)
		{
			//see if this character has a substitution
			search_i = index(state->lsys->vars, *c);
			if (search_i)
			{
				//find the index into the proposition array
				size_t i = search_i - state->lsys->vars;
				//replace current string with propositiotn
				char *temp = state->str;
				state->str = state->lsys->productions[i];

				//process the new string
				++state->depth;
				lsystem_draw_r(state);
				--state->depth;
				
				//restore and move on
				state->str = temp;
				++c;
				continue;
			}
		}

		//If at the maximum recursion depth, or the character
		//was not substituted then process it
		switch (*c)
		{
			case LSYSTEM_LINE :
				draw_line(state);
			break;
			case LSYSTEM_MOVE :
				(void) c;

				int dx, dy;
				find_slope(state, &dx, &dy);
				state->x += dx;
				state->y += dy;
			break;
			case LSYSTEM_ROTATE_P :
				state->angle += state->lsys->angle*negate;
			break;	
			case LSYSTEM_ROTATE_N :
				state->angle -= state->lsys->angle*negate;
			break;	
			case LSYSTEM_SAVE:
				(void) c;

				//copy current state
				struct lsystem_state new_state;
				memcpy(&new_state, state, sizeof(new_state));
				new_state.str = c + 1;
				
				//call recursive
				lsystem_draw_r(&new_state);
				
				//advance past save area
				search_i = index(c, LSYSTEM_RESTORE);
				if (search_i)
					c = search_i;
				else
					//DPRINT("Missing close of save state at: %s\n", c);
			break;
			case LSYSTEM_RESTORE :
				done = true;
			break;
			case LSYSTEM_ANGLE :
				angle = true;
			case LSYSTEM_SCALE :
				++c;
				
				float amount;
				int len;
				int r = sscanf(c, "%f%n", &amount, &len);
				if (r != EOF && len > 0)
				{
					if (angle)
						state->angle = amount;
					else
						state->len *= amount;
					//advance pointer to end of number
					c += len - 1;
				}
				else
				{
					//DPRINT("Invaild number at: %s\n", c);
				}
			break;
			case LSYSTEM_REVERSE :
				negate *= -1;
			break;
			case LSYSTEM_COLOUR :
				if (state->colour_mode == LSYSTEM_COLOUR_INLINE)
				{
					state->ci = (state->ci + 1) % state->num_colours;
					state->colour[0] = state->colour_list[state->ci * 3];
					state->colour[1] = state->colour_list[state->ci * 3 + 1];
					state->colour[2] = state->colour_list[state->ci * 3 + 2];
				}
			break;
			case LSYSTEM_TOP :
				state->y = state->h - 1;
			break;
			case LSYSTEM_BOTTOM :
				state->y = 0;
			break;
			case LSYSTEM_LEFT :
				state->x = 0;
			break;
			case LSYSTEM_RIGHT :
				state->x = state->w - 1;
			break;
			case LSYSTEM_CENTER_H :
				state->x = state->w/2;
			break;
			case LSYSTEM_CENTER_V :
				state->y = state->h/2;
			break;
			default : break;
		}
		
		//next character
		++c;
	}

}

void *lsystem_draw_colour(struct lsystem *l, size_t w, size_t h,
						  enum lsystem_colour_mode cm,
						  unsigned char *colours,
						  int num_colours
						  )
{
	void *data = malloc(w * h * 3);
	//memset(data, 180, w*h*3);
	
	unsigned char background[3] = {255, 255, 255};
	if (colours)
	{
		background[0] = colours[0];
		background[1] = colours[1];
		background[2] = colours[2];
		colours += 3;
	}
	size_t i = 0;
	char *d = data;
	for (i = 0; i < w*h; i++)
	{
		d[0] = background[0];
		d[1] = background[1];
		d[2] = background[2];
		d+= 3;
	}

	struct lsystem_state state;
	state.w = w;
	state.h = h;
	state.x = w/2;
	state.y = h/2;
	
	state.depth = 0;

	state.lsys = l;
	state.angle = 0.0;
	state.len = l->len;
	state.str = l->axiom;

	state.colour_mode = cm;
	state.colour_list = colours;
	state.ci = 0;
	if (colours)
	{
		state.colour[0] = colours[0];
		state.colour[1] = colours[1];
		state.colour[2] = colours[2];
	}
	else
	{
		state.colour[0] = 0;
		state.colour[1] = 0;
		state.colour[2] = 0;
	}
	state.num_colours = num_colours;

	state.data = (char *) data;

	lsystem_draw_r(&state);

	return data;
}

void *lsystem_draw(struct lsystem *l, size_t w, size_t h)
{
	return lsystem_draw_colour(l, w, h, LSYSTEM_COLOUR_NONE, NULL, 0);
}

/** Default background red */
#define BR 60
/** Default background green */
#define BG 120
/** Default background blue */
#define BB 200
void *lsystem_test(unsigned int id, float len, size_t w, size_t h)
{
	struct lsystem l;
	memcpy(&l, &lsystem_tests[id], sizeof(l));
	
	l.len += len;
	
	enum lsystem_colour_mode mode = LSYSTEM_COLOUR_NONE;
	unsigned char *colours = NULL;
	int num_colours = 0;
	switch (id)
	{
		case 0:
		case 1:
		case 2:
			mode = LSYSTEM_COLOUR_DEPTH;
			if (id == 2)
				mode = LSYSTEM_COLOUR_INLINE;
			unsigned char c1[] = {
								BR, BG, BB,
								110, 70, 0, 
								15, 130, 10,
								};
			colours = c1;
			num_colours = 2;
		break;
		case 3:
			mode = LSYSTEM_COLOUR_LINE;
			unsigned char c3[] = {
								BR, BG, BB,
								150, 0, 150, 
								0, 0, 240, 
								180, 0, 40, 
								0, 0, 240, 
								150, 0, 150, 
								0, 0, 240, 
								150, 0, 150, 
								};
			colours = c3;
			num_colours = 7;
		break;
		case 4:
			mode = LSYSTEM_COLOUR_LINE;
			unsigned char c4[] = {
								   BR, BG, BB,
								   0, 0, 0,
								 };
			colours = c4;
			num_colours = 2;
			if (id == 4) num_colours = 1;
		break;	
		case 5:
		case 6:
			mode = LSYSTEM_COLOUR_INLINE;
			unsigned char c6[] = {
									BR, BG, BB,
									0, 0, 240,
									0, 0, 240,
									0, 0, 240,
									180, 0, 40,
									180, 0, 40,
									180, 0, 40,
									140, 0, 140,
									140, 0, 140,
									140, 0, 140,
									180, 0, 40,
									180, 0, 40,
									180, 0, 40,
								};
			colours = c6;
			num_colours = 9;
		break;
		case 7:
			mode = LSYSTEM_COLOUR_LINE;
			unsigned char c7[] = {
								   BR, BG, BB,
								   240, 0, 0,
								   205, 205, 0,
								 };
			colours = c7;
			num_colours = 2;
			if (id == 4) num_colours = 1;
		break;
	}
	
	return lsystem_draw_colour(&l, w, h, mode, colours, num_colours);
}
