/* Matthew Phillips mephilli 20066500 CS488 A5
 *
 * Copyright (c) 2005 Matthew Phillips
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/** @file lsystem.h 
 *
 *  An lsystem parser
 */
#ifndef LSYSTEM_H
#define LSYSTEM_H

//These are fairly standard lsystem symbols

/** Draws a line */
#define LSYSTEM_LINE 'F'
/** Moves without drawing a line */
#define LSYSTEM_MOVE 'M'
/** Turns left (positive)*/
#define LSYSTEM_ROTATE_P '+'
/** Turns right (negative) */
#define LSYSTEM_ROTATE_N '-'
/** Save the current state */
#define LSYSTEM_SAVE '['
/** Restore the previously saved state */
#define LSYSTEM_RESTORE ']'

//These are some addiotns I have seen
/** Change the size of the line */
#define LSYSTEM_SCALE 'S'
/** Reverse meaning of +,- */
#define LSYSTEM_REVERSE 'N'

//These are additions I made up
/** Switch to next colour */
#define LSYSTEM_COLOUR 'C'
/** Move to top of data area */
#define LSYSTEM_TOP	'T'
/** Move to bottom of data area */
#define LSYSTEM_BOTTOM 'B'
/** Move to left of data area */
#define LSYSTEM_LEFT 'L'
/** Move to right of data area */
#define LSYSTEM_RIGHT 'R'
/** Move to horizonatl center of data area */
#define LSYSTEM_CENTER_H 'H'
/** Move to vertical center of data area */
#define LSYSTEM_CENTER_V 'V'
/** Set the current angle */
#define LSYSTEM_ANGLE 'A'

/** The possible colour modes */
enum lsystem_colour_mode
{
	LSYSTEM_COLOUR_NONE,
	//Colours are divided evenly by depth
	LSYSTEM_COLOUR_DEPTH,
	//Colour is changed each time a line is drawn
	LSYSTEM_COLOUR_LINE,
	//Colour is changing using C command in lsystem string
	LSYSTEM_COLOUR_INLINE
};

/** Struct which defines an lsystem */
struct lsystem
{
	/** The starting string */
	char *axiom;
	/** List the variables used. Variables should be lower case letters */
	char *vars;
	/** List of productions.
	 *
	 * There should be one production for each variable, and they should
	 * occur in the same order as in the variable list.
	 */
	char *productions[5];

	/** Recursion depth */
	unsigned int depth;
	/** What angle to turn at for each turn command */
	float angle;
	/** The default line size */
	float len;
};

/** The number of predefined lsystems, used for testing */
#define NUM_LSYSTEM_TESTS 8
/** Predefine lsystems */
static const struct lsystem lsystem_tests[NUM_LSYSTEM_TESTS] =
{
	//Tree **
	{"HBA90FFy", "xy", 
		{"F[S0.5+++++++++x]-F[S0.4-----------Nx]S0.6x",
		 "F[A160S0.4x]F[A10S0.5x]x}"
		},
		9, 7, 50},
	//Bush **
	{"HBA90F", "F", {"FF-[-F+F+F]+[+F-F-F]"}, 4, 22.5, 8},
	//Thistle **
	{"VLyA180Ry", "xy", {"FxC[-Fx][-Fx][+Fx]Fx", "[++FFx][--FFx]"}, 5, 30, 4},
	//Square **
	{"[F]+[F]+[F]+[F]", "F", {"F[+F][-F]F",}, 5, 90, 6},
	//Circle **
	{"HVxA180x", "x", {"x+F",}, 360, 1, 5},
	//Kosh Island **
	{"[x]+[x]+[x]+[x]", "Fx", {"F+CF-F-FCF+F+F-F", "F+F+F+F"}, 4, 90, 2},
	//Kosh2? **
	{"BL[F]TRA180F", "F", {"F+CF-F-FC+F",}, 5, 90, 5},
	//Dragon Curve **
	{"[-x]+++x", "xy", {"x+F+y", "x-F-y"}, 16, 45, 1},
	//{"x", "x", {"F+FS1.1+FxF+FS1.1+Fx"}, 4, 20, 1},
};

/** String representation of lsystem tests */
static const char *lsystem_test_s[NUM_LSYSTEM_TESTS] = 
{
	"Tree   ",
	"Bush   ",
	"Thistle",
	"Square ",
	"Circle ",
	"Kosh   ",
	"Kosh2  ",
	"Dragon ",
};


/** Draws an lsystem with colour
 * 
 *  @param l The lsystem to draw
 *  @param w The width
 *  @param h The height
 *  @param cm The colour mode
 *  @param colours A list of colours (groups of 3 used for colours)
 *  The first colour should be the desired background colour, it should
 *  not be included in the count of num_colours
 *  @param num_colours The number of colours in the list
 *  
 *  @return Returns a pointer to a w*h*3 chunk of memory which holds the
 *  lsystem
 * */
void *lsystem_draw_colour(struct lsystem *l, size_t w, size_t h,
							enum lsystem_colour_mode cm,
							unsigned char *colours,
							int num_colours);

/** Draws an lsystem
 * 
 *  @param l The lsystem to draw
 *  @param w The width
 *  @param h The height
 *  
 *  @return Returns a pointer to a w*h*3 chunk of memory which holds the
 *  lsystem
 * */
void *lsystem_draw(struct lsystem *l, size_t w, size_t h);

/** Draws an lsystem test 
 *  @param id The test number
 *  @param len Line length
 *  @param w The width
 *  @param h The height
 */
void *lsystem_test(unsigned int id, float len, size_t w, size_t h);

#endif /* GL_PRIMITIVE_H */
