/* Matthew Phillips mephilli 20066500 CS488 A5
 *
 * Copyright (c) 2005 Matthew Phillips
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/** @file psystem.c 
 *
 *  Responsible for particle system.
 * 
 */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <errno.h>

#include "psystem.h"
#include "glprimitive.h"

/** The average spin */
#define SPIN_MEAN 15
/** The spin variance */
#define SPIN_VAR  5
/** The average particle life */
#define LIFE_MEAN ((psys->len/p->dir[2])*0.8)
/** The particle life variance */
#define LIFE_VAR 0.4
/** The average number of particels generated per frame */
#define MEAN_GEN (MAX_NUM_PARTICLES)
/** The particle generation variance */
#define VAR_GEN (MAX_NUM_PARTICLES/2)


/** Returns a random number in the range [0..max) 
 *  @param max The maximum value
 */
static GLdouble rand1(GLdouble max)
{
	return (max*rand())/(RAND_MAX+1.0);
}

/** Returns a number close to the given mean
 *  @param mean The mean value
 *  @param var The maximum variance from the mean
 */
static GLdouble rand2(GLdouble mean, GLdouble var)
{
	return mean + (2*var*rand())/(RAND_MAX+1.0) - var;
}

/** Generates a new particle 
 *  @param psys A pointer to the particle system 
 */
void
new_particle(struct psystem *psys)
{
	struct psystem_particle *p = &psys->particles[psys->num_particles++];

	p->size = rand2(psys->radius/6.0, psys->radius/8.0);
	
	//pick a randum place on the circle
	GLdouble angle = rand1(360);
	GLdouble r = rand2(psys->radius/2.0, psys->radius/2.0);
	p->pos[0] = r*cos(angle*M_PI/180);
	p->pos[1] = r*sin(angle*M_PI/180); 
	p->pos[2] = 0;

	//The point at the end of the path
	GLdouble r2 = rand2(psys->radius2/2.0, psys->radius2/2.0);
	GLdouble x2 = r2*cos(angle*M_PI/180);
	GLdouble y2 = r2*sin(angle*M_PI/180);
	  //Direction 
	p->dir[0] = x2 - p->pos[0];
	p->dir[1] = y2 - p->pos[1]; 
	p->dir[2] = psys->len;
	  //Normalize then multiply by a speed
	GLdouble n=p->dir[0]*p->dir[0] + p->dir[1]*p->dir[1] + p->dir[2]*p->dir[2];
	n = rand2(psys->len/4.0, psys->len/8.0)/sqrt(n);
	p->dir[0] *= n;
	p->dir[1] *= n;
	p->dir[2] *= n;

	//pick a random spin
	p->spin[0] = rand2(SPIN_MEAN, SPIN_VAR);
	p->spin[1] = rand2(SPIN_MEAN, SPIN_VAR);
	p->spin[2] = rand2(SPIN_MEAN, SPIN_VAR);
	p->rot[0] = p->spin[0] * p->spin[1];
	p->rot[1] = p->spin[1] * p->spin[2];
	p->rot[2] = p->spin[2] * p->spin[0];

	p->life = (unsigned int) rand2(LIFE_MEAN, LIFE_VAR);
}

/** Adances the given particel 
 *  @param p The particle to advance
 *  @return Returns true if the particle is still alive. false otherwise
 */
static bool
advance_particle(struct psystem_particle *p)
{
	size_t i;
	for (i = 0; i < 3; i++)
	{
		p->pos[i] += p->dir[i];
		p->rot[i] += p->spin[i];
		if (p->rot[i] > 360.0)
			p->rot[i] = 0.0;
	}

	if (p->life > 0)
	{
		--p->life;
		return true;
	}
	else
		return false;
}

void
psystem_init(struct psystem *psys, GLdouble r1, GLdouble r2, GLdouble len,
			 enum psystem_colour c)
{
	static bool first_time = true;
	
	psys->radius = r1;
	psys->radius2 = r2;
	psys->len = len;
	psys->num_particles = 0;
	psys->colour = c;
}

void
psystem_reset(struct psystem *psys)
{
	psys->num_particles = 0;
}

void
psystem_advance(struct psystem *psys)
{
	//used so only the pre existing particles will be advanced
	size_t orig_size = psys->num_particles;
	
	//generate a new particle
	int num = (int) rand2(MEAN_GEN, VAR_GEN);
	while (num > 0 && psys->num_particles < MAX_NUM_PARTICLES)
	{
		new_particle(psys);
		--num;
	}
	
	//advance particles
	size_t i = 0;
	while (i < orig_size && i < psys->num_particles)
	{
		if (advance_particle(&psys->particles[i]))
			++i;
		else
			psys->particles[i] = psys->particles[psys->num_particles--];
	}
}

void
psystem_draw(struct psystem *psys)
{
	static const GLfloat spec_w[] = {0.1, 0.1, 0.1, 1.0};
	static const GLfloat shinny_w = 1;
	glMaterialfv(GL_FRONT, GL_SPECULAR, spec_w);
	glMaterialf(GL_FRONT, GL_SHININESS, shinny_w);
									 
	size_t i = 0;
	while (i < psys->num_particles)
	{
		struct psystem_particle *p = &psys->particles[i];

		//determine colour
		GLdouble t = p->pos[2]/psys->len;
		GLfloat colour[4];
		switch (psys->colour)
		{
			case PSYSTEM_COLOUR_ORANGE: 
				colour[0] = 1;
				colour[1] = 0.3;
				colour[2] = 0.3;
				colour[3] = 1.0;
			break;
			case PSYSTEM_COLOUR_BLUE:
				colour[0] = 1 - 1*sqrt(t*t);
				colour[1] = 1 - 1*sqrt(t*t);
				colour[2] = 1;
				colour[3] = 1.0;
			break;
			case PSYSTEM_COLOUR_RAINBOW:
				t *= 2.0;
				if (t < 1.0)
				{
					colour[0] = 0 + 1*sqrt(t*t);
					colour[1] = 1 - 1*sqrt(t*t);
					colour[2] = 0;
				}
				else
				{
					t -= 1;
					colour[0] = 1*sqrt(t);
					colour[1] = 0;
					colour[2] = 1*sqrt(t*t);
				}
				colour[3] = 1.0;
			break;
			case PSYSTEM_COLOUR_RANDOM:
				colour[0] = rand1(1.0);
				colour[1] = rand1(1.0);
				colour[2] = rand1(1.0);
				colour[3] = 1.0;
			break;
		}
		glMaterialfv(GL_FRONT, GL_DIFFUSE, colour);
		
		glPushMatrix();
			//move
			glTranslated(p->pos[0], p->pos[1], p->pos[2]);
			//spin
			glRotatef(p->rot[0], 1.0, 0.0, 0.0);
			glRotatef(p->rot[1], 0.0, 1.0, 0.0);
			glRotatef(p->rot[2], 0.0, 0.0, 1.0);
			//draw
			glprimitive_box(-p->size/2.0, -p->size/2.0, -p->size/2.0,
							p->size, p->size, p->size);
		glPopMatrix();
		++i;
	}
}
